# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .claim_atom import ClaimAtom
from .simple_payment_result import SimplePaymentResult

__all__ = ["PathPaymentStrictSendResultSuccess"]


class PathPaymentStrictSendResultSuccess:
    """
    XDR Source Code::

        struct
            {
                ClaimAtom offers<>;
                SimplePaymentResult last;
            }
    """

    def __init__(
        self,
        offers: List[ClaimAtom],
        last: SimplePaymentResult,
    ) -> None:
        _expect_max_length = 4294967295
        if offers and len(offers) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `offers` should be {_expect_max_length}, but got {len(offers)}."
            )
        self.offers = offers
        self.last = last

    def pack(self, packer: Packer) -> None:
        packer.pack_uint(len(self.offers))
        for offers_item in self.offers:
            offers_item.pack(packer)
        self.last.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> PathPaymentStrictSendResultSuccess:
        length = unpacker.unpack_uint()
        offers = []
        for _ in range(length):
            offers.append(ClaimAtom.unpack(unpacker))
        last = SimplePaymentResult.unpack(unpacker)
        return cls(
            offers=offers,
            last=last,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> PathPaymentStrictSendResultSuccess:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> PathPaymentStrictSendResultSuccess:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.offers,
                self.last,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.offers == other.offers and self.last == other.last

    def __repr__(self):
        out = [
            f"offers={self.offers}",
            f"last={self.last}",
        ]
        return f"<PathPaymentStrictSendResultSuccess [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .int64 import Int64
from .uint32 import Uint32

__all__ = ["ConfigSettingContractBandwidthV0"]


class ConfigSettingContractBandwidthV0:
    """
    XDR Source Code::

        struct ConfigSettingContractBandwidthV0
        {
            // Maximum sum of all transaction sizes in the ledger in bytes
            uint32 ledgerMaxTxsSizeBytes;
            // Maximum size in bytes for a transaction
            uint32 txMaxSizeBytes;

            // Fee for 1 KB of transaction size
            int64 feeTxSize1KB;
        };
    """

    def __init__(
        self,
        ledger_max_txs_size_bytes: Uint32,
        tx_max_size_bytes: Uint32,
        fee_tx_size1_kb: Int64,
    ) -> None:
        self.ledger_max_txs_size_bytes = ledger_max_txs_size_bytes
        self.tx_max_size_bytes = tx_max_size_bytes
        self.fee_tx_size1_kb = fee_tx_size1_kb

    def pack(self, packer: Packer) -> None:
        self.ledger_max_txs_size_bytes.pack(packer)
        self.tx_max_size_bytes.pack(packer)
        self.fee_tx_size1_kb.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ConfigSettingContractBandwidthV0:
        ledger_max_txs_size_bytes = Uint32.unpack(unpacker)
        tx_max_size_bytes = Uint32.unpack(unpacker)
        fee_tx_size1_kb = Int64.unpack(unpacker)
        return cls(
            ledger_max_txs_size_bytes=ledger_max_txs_size_bytes,
            tx_max_size_bytes=tx_max_size_bytes,
            fee_tx_size1_kb=fee_tx_size1_kb,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ConfigSettingContractBandwidthV0:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ConfigSettingContractBandwidthV0:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.ledger_max_txs_size_bytes,
                self.tx_max_size_bytes,
                self.fee_tx_size1_kb,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.ledger_max_txs_size_bytes == other.ledger_max_txs_size_bytes
            and self.tx_max_size_bytes == other.tx_max_size_bytes
            and self.fee_tx_size1_kb == other.fee_tx_size1_kb
        )

    def __repr__(self):
        out = [
            f"ledger_max_txs_size_bytes={self.ledger_max_txs_size_bytes}",
            f"tx_max_size_bytes={self.tx_max_size_bytes}",
            f"fee_tx_size1_kb={self.fee_tx_size1_kb}",
        ]
        return f"<ConfigSettingContractBandwidthV0 [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .claim_atom import ClaimAtom
from .manage_offer_success_result_offer import ManageOfferSuccessResultOffer

__all__ = ["ManageOfferSuccessResult"]


class ManageOfferSuccessResult:
    """
    XDR Source Code::

        struct ManageOfferSuccessResult
        {
            // offers that got claimed while creating this offer
            ClaimAtom offersClaimed<>;

            union switch (ManageOfferEffect effect)
            {
            case MANAGE_OFFER_CREATED:
            case MANAGE_OFFER_UPDATED:
                OfferEntry offer;
            case MANAGE_OFFER_DELETED:
                void;
            }
            offer;
        };
    """

    def __init__(
        self,
        offers_claimed: List[ClaimAtom],
        offer: ManageOfferSuccessResultOffer,
    ) -> None:
        _expect_max_length = 4294967295
        if offers_claimed and len(offers_claimed) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `offers_claimed` should be {_expect_max_length}, but got {len(offers_claimed)}."
            )
        self.offers_claimed = offers_claimed
        self.offer = offer

    def pack(self, packer: Packer) -> None:
        packer.pack_uint(len(self.offers_claimed))
        for offers_claimed_item in self.offers_claimed:
            offers_claimed_item.pack(packer)
        self.offer.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ManageOfferSuccessResult:
        length = unpacker.unpack_uint()
        offers_claimed = []
        for _ in range(length):
            offers_claimed.append(ClaimAtom.unpack(unpacker))
        offer = ManageOfferSuccessResultOffer.unpack(unpacker)
        return cls(
            offers_claimed=offers_claimed,
            offer=offer,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ManageOfferSuccessResult:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ManageOfferSuccessResult:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.offers_claimed,
                self.offer,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.offers_claimed == other.offers_claimed and self.offer == other.offer

    def __repr__(self):
        out = [
            f"offers_claimed={self.offers_claimed}",
            f"offer={self.offer}",
        ]
        return f"<ManageOfferSuccessResult [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .archival_proof_type import ArchivalProofType
from .existence_proof_body import ExistenceProofBody
from .nonexistence_proof_body import NonexistenceProofBody

__all__ = ["ArchivalProofBody"]


class ArchivalProofBody:
    """
    XDR Source Code::

        union switch (ArchivalProofType t)
            {
            case EXISTENCE:
                NonexistenceProofBody nonexistenceProof;
            case NONEXISTENCE:
                ExistenceProofBody existenceProof;
            }
    """

    def __init__(
        self,
        t: ArchivalProofType,
        nonexistence_proof: NonexistenceProofBody = None,
        existence_proof: ExistenceProofBody = None,
    ) -> None:
        self.t = t
        self.nonexistence_proof = nonexistence_proof
        self.existence_proof = existence_proof

    def pack(self, packer: Packer) -> None:
        self.t.pack(packer)
        if self.t == ArchivalProofType.EXISTENCE:
            if self.nonexistence_proof is None:
                raise ValueError("nonexistence_proof should not be None.")
            self.nonexistence_proof.pack(packer)
            return
        if self.t == ArchivalProofType.NONEXISTENCE:
            if self.existence_proof is None:
                raise ValueError("existence_proof should not be None.")
            self.existence_proof.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ArchivalProofBody:
        t = ArchivalProofType.unpack(unpacker)
        if t == ArchivalProofType.EXISTENCE:
            nonexistence_proof = NonexistenceProofBody.unpack(unpacker)
            return cls(t=t, nonexistence_proof=nonexistence_proof)
        if t == ArchivalProofType.NONEXISTENCE:
            existence_proof = ExistenceProofBody.unpack(unpacker)
            return cls(t=t, existence_proof=existence_proof)
        return cls(t=t)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ArchivalProofBody:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ArchivalProofBody:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.t,
                self.nonexistence_proof,
                self.existence_proof,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.t == other.t
            and self.nonexistence_proof == other.nonexistence_proof
            and self.existence_proof == other.existence_proof
        )

    def __repr__(self):
        out = []
        out.append(f"t={self.t}")
        (
            out.append(f"nonexistence_proof={self.nonexistence_proof}")
            if self.nonexistence_proof is not None
            else None
        )
        (
            out.append(f"existence_proof={self.existence_proof}")
            if self.existence_proof is not None
            else None
        )
        return f"<ArchivalProofBody [{', '.join(out)}]>"



from decimal import Decimal
from typing import List, Union

from ...asset import Asset
from ...call_builder.base.base_call_builder import BaseCallBuilder
from ...utils import convert_assets_to_horizon_param

__all__ = ["BaseStrictSendPathsCallBuilder"]


class BaseStrictSendPathsCallBuilder(BaseCallBuilder):
    """Creates a new :class:`StrictSendPathsCallBuilder` pointed to server defined by horizon_url.

    The Stellar Network allows payments to be made across assets through path
    payments. A strict send path payment specifies a series of assets to route a
    payment through, from source asset (the asset debited from the payer) to
    destination asset (the asset credited to the payee).

    A strict send path search is specified using:

    - The source asset
    - The source amount
    - The destination assets or destination account.

    As part of the search, horizon will load a list of assets available to the
    source address and will find any payment paths from those source assets to
    the desired destination asset. The search's source_amount parameter will be
    used to determine if there a given path can satisfy a payment of the desired
    amount.

    See `List Strict Send Payment Paths <https://developers.stellar.org/api/aggregations/paths/strict-send/>`__ for more information.

    :param source_asset: The asset to be sent.
    :param source_amount: The amount, denominated in the source asset, that any returned path should be able to satisfy.
    :param destination: The destination account or the destination assets.
    :param horizon_url: Horizon server URL.
    """

    def __init__(
        self,
        source_asset: Asset,
        source_amount: Union[str, Decimal],
        destination: Union[str, List[Asset]],
        **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self.endpoint: str = "paths/strict-send"

        if isinstance(source_amount, Decimal):
            source_amount = str(source_amount)

        params = {
            "source_amount": source_amount,
            "source_asset_type": source_asset.type,
            "source_asset_code": (
                None if source_asset.is_native() else source_asset.code
            ),
            "source_asset_issuer": source_asset.issuer,
        }

        if isinstance(destination, str):
            params["destination_account"] = destination
        else:
            params["destination_assets"] = convert_assets_to_horizon_param(destination)

        self._add_query_params(params)



from . import xdr as stellar_xdr
from .asset import Asset
from .utils import sha256

__all__ = ["LiquidityPoolAsset", "LIQUIDITY_POOL_FEE_V18"]

#: LIQUIDITY_POOL_FEE_V18 is the default liquidity pool fee in protocol v18.
#: It defaults to 30 base points (0.3%).
LIQUIDITY_POOL_FEE_V18 = stellar_xdr.LIQUIDITY_POOL_FEE_V18


class LiquidityPoolAsset:
    """The :class:`LiquidityPoolAsset` object, which represents a liquidity pool trustline change.

    :param asset_a: The first asset in the Pool, it must respect the rule asset_a < asset_b.
        See :func:`stellar_sdk.liquidity_pool_asset.LiquidityPoolAsset.is_valid_lexicographic_order`
        for more details on how assets are sorted.
    :param asset_b: The second asset in the Pool, it must respect the rule asset_a < asset_b.
        See :func:`stellar_sdk.liquidity_pool_asset.LiquidityPoolAsset.is_valid_lexicographic_order`
        for more details on how assets are sorted.
    :param fee: The liquidity pool fee. For now the only fee supported is `30`.
    :raise: :exc:`ValueError <stellar_sdk.exceptions.ValueError>`
    """

    def __init__(
        self, asset_a: Asset, asset_b: Asset, fee: int = LIQUIDITY_POOL_FEE_V18
    ) -> None:
        if not self.is_valid_lexicographic_order(asset_a, asset_b):
            raise ValueError("`Assets are not in lexicographic order.")

        if fee != LIQUIDITY_POOL_FEE_V18:
            raise ValueError("`fee` is invalid.")

        self.type: str = "liquidity_pool_shares"
        self.asset_a: Asset = asset_a
        self.asset_b: Asset = asset_b
        self.fee: int = fee

    @property
    def liquidity_pool_id(self) -> str:
        """Computes the liquidity pool id for current instance.

        :return: Liquidity pool id.
        """
        return self._liquidity_pool_id_bytes.hex()

    @staticmethod
    def is_valid_lexicographic_order(asset_a: Asset, asset_b: Asset) -> bool:
        """Compares if asset_a < asset_b according with the criteria:

        1. First compare the type (eg. native before alphanum4 before alphanum12).
        2. If the types are equal, compare the assets codes.
        3. If the asset codes are equal, compare the issuers.

        :param asset_a: The first asset in the lexicographic order.
        :param asset_b: The second asset in the lexicographic order.
        :return: return `True` if asset_a < asset_b
        """
        if asset_a == asset_b:
            return False

        # Compare asset types.
        if asset_a.type == "native":
            return True
        elif asset_a.type == "credit_alphanum4":
            if asset_b.type == "native":
                return False
            if asset_b.type == "credit_alphanum12":
                return True
        else:
            # asset_a.type == "credit_alphanum12"
            if asset_b.type != "credit_alphanum12":
                return False

        # Compare asset codes.
        if asset_a.code != asset_b.code:
            assert asset_a.code is not None
            assert asset_b.code is not None
            return asset_a.code < asset_b.code

        # Compare asset issuer.
        assert asset_a.issuer is not None
        assert asset_b.issuer is not None
        return asset_a.issuer < asset_b.issuer

    def to_change_trust_asset_xdr_object(self) -> stellar_xdr.ChangeTrustAsset:
        """Returns the xdr object for this ChangeTrustAsset object.

        :return: XDR ChangeTrustAsset object
        """
        liquidity_pool_parameters = self._liquidity_pool_parameters()
        return stellar_xdr.ChangeTrustAsset(
            stellar_xdr.AssetType.ASSET_TYPE_POOL_SHARE,
            liquidity_pool=liquidity_pool_parameters,
        )

    @classmethod
    def from_xdr_object(
        cls, xdr_object: stellar_xdr.ChangeTrustAsset
    ) -> "LiquidityPoolAsset":
        """Create a :class:`LiquidityPoolAsset` from an XDR ChangeTrustAsset object.

        :param xdr_object: The XDR ChangeTrustAsset object.
        :return: A new :class:`LiquidityPoolAsset` object from the given XDR ChangeTrustAsset object.
        """
        asset_type = xdr_object.type
        if asset_type == stellar_xdr.AssetType.ASSET_TYPE_POOL_SHARE:
            assert xdr_object.liquidity_pool is not None
            assert xdr_object.liquidity_pool.constant_product is not None
            asset_a = Asset.from_xdr_object(
                xdr_object.liquidity_pool.constant_product.asset_a
            )
            asset_b = Asset.from_xdr_object(
                xdr_object.liquidity_pool.constant_product.asset_b
            )
            fee = xdr_object.liquidity_pool.constant_product.fee.int32
            return cls(asset_a, asset_b, fee)
        else:
            raise ValueError(f"Unexpected asset type: {asset_type}")

    @property
    def _liquidity_pool_id_bytes(self) -> bytes:
        liquidity_pool_parameters = self._liquidity_pool_parameters()
        return sha256(liquidity_pool_parameters.to_xdr_bytes())

    def _liquidity_pool_parameters(self) -> stellar_xdr.LiquidityPoolParameters:
        liquidity_pool_constant_product_parameters = (
            stellar_xdr.LiquidityPoolConstantProductParameters(
                self.asset_a.to_xdr_object(),
                self.asset_b.to_xdr_object(),
                stellar_xdr.Int32(self.fee),
            )
        )
        return stellar_xdr.LiquidityPoolParameters(
            stellar_xdr.LiquidityPoolType.LIQUIDITY_POOL_CONSTANT_PRODUCT,
            liquidity_pool_constant_product_parameters,
        )

    def __hash__(self):
        return hash((self.asset_a, self.asset_b, self.fee, self.type))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.asset_a == other.asset_a
            and self.asset_b == other.asset_b
            and self.fee == other.fee
        )

    def __repr__(self):
        return f"<LiquidityPoolAsset [asset_a={self.asset_a}, asset_b={self.asset_b}, fee={self.fee}, type={self.type}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .ledger_entry_changes import LedgerEntryChanges
from .ledger_upgrade import LedgerUpgrade

__all__ = ["UpgradeEntryMeta"]


class UpgradeEntryMeta:
    """
    XDR Source Code::

        struct UpgradeEntryMeta
        {
            LedgerUpgrade upgrade;
            LedgerEntryChanges changes;
        };
    """

    def __init__(
        self,
        upgrade: LedgerUpgrade,
        changes: LedgerEntryChanges,
    ) -> None:
        self.upgrade = upgrade
        self.changes = changes

    def pack(self, packer: Packer) -> None:
        self.upgrade.pack(packer)
        self.changes.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> UpgradeEntryMeta:
        upgrade = LedgerUpgrade.unpack(unpacker)
        changes = LedgerEntryChanges.unpack(unpacker)
        return cls(
            upgrade=upgrade,
            changes=changes,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> UpgradeEntryMeta:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> UpgradeEntryMeta:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.upgrade,
                self.changes,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.upgrade == other.upgrade and self.changes == other.changes

    def __repr__(self):
        out = [
            f"upgrade={self.upgrade}",
            f"changes={self.changes}",
        ]
        return f"<UpgradeEntryMeta [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .transaction_history_result_entry_ext import TransactionHistoryResultEntryExt
from .transaction_result_set import TransactionResultSet
from .uint32 import Uint32

__all__ = ["TransactionHistoryResultEntry"]


class TransactionHistoryResultEntry:
    """
    XDR Source Code::

        struct TransactionHistoryResultEntry
        {
            uint32 ledgerSeq;
            TransactionResultSet txResultSet;

            // reserved for future use
            union switch (int v)
            {
            case 0:
                void;
            }
            ext;
        };
    """

    def __init__(
        self,
        ledger_seq: Uint32,
        tx_result_set: TransactionResultSet,
        ext: TransactionHistoryResultEntryExt,
    ) -> None:
        self.ledger_seq = ledger_seq
        self.tx_result_set = tx_result_set
        self.ext = ext

    def pack(self, packer: Packer) -> None:
        self.ledger_seq.pack(packer)
        self.tx_result_set.pack(packer)
        self.ext.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> TransactionHistoryResultEntry:
        ledger_seq = Uint32.unpack(unpacker)
        tx_result_set = TransactionResultSet.unpack(unpacker)
        ext = TransactionHistoryResultEntryExt.unpack(unpacker)
        return cls(
            ledger_seq=ledger_seq,
            tx_result_set=tx_result_set,
            ext=ext,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> TransactionHistoryResultEntry:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> TransactionHistoryResultEntry:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.ledger_seq,
                self.tx_result_set,
                self.ext,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.ledger_seq == other.ledger_seq
            and self.tx_result_set == other.tx_result_set
            and self.ext == other.ext
        )

    def __repr__(self):
        out = [
            f"ledger_seq={self.ledger_seq}",
            f"tx_result_set={self.tx_result_set}",
            f"ext={self.ext}",
        ]
        return f"<TransactionHistoryResultEntry [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .hmac_sha256_mac import HmacSha256Mac
from .stellar_message import StellarMessage
from .uint64 import Uint64

__all__ = ["AuthenticatedMessageV0"]


class AuthenticatedMessageV0:
    """
    XDR Source Code::

        struct
            {
                uint64 sequence;
                StellarMessage message;
                HmacSha256Mac mac;
            }
    """

    def __init__(
        self,
        sequence: Uint64,
        message: StellarMessage,
        mac: HmacSha256Mac,
    ) -> None:
        self.sequence = sequence
        self.message = message
        self.mac = mac

    def pack(self, packer: Packer) -> None:
        self.sequence.pack(packer)
        self.message.pack(packer)
        self.mac.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> AuthenticatedMessageV0:
        sequence = Uint64.unpack(unpacker)
        message = StellarMessage.unpack(unpacker)
        mac = HmacSha256Mac.unpack(unpacker)
        return cls(
            sequence=sequence,
            message=message,
            mac=mac,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> AuthenticatedMessageV0:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> AuthenticatedMessageV0:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.sequence,
                self.message,
                self.mac,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.sequence == other.sequence
            and self.message == other.message
            and self.mac == other.mac
        )

    def __repr__(self):
        out = [
            f"sequence={self.sequence}",
            f"message={self.message}",
            f"mac={self.mac}",
        ]
        return f"<AuthenticatedMessageV0 [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .hash import Hash
from .ledger_header import LedgerHeader
from .ledger_header_history_entry_ext import LedgerHeaderHistoryEntryExt

__all__ = ["LedgerHeaderHistoryEntry"]


class LedgerHeaderHistoryEntry:
    """
    XDR Source Code::

        struct LedgerHeaderHistoryEntry
        {
            Hash hash;
            LedgerHeader header;

            // reserved for future use
            union switch (int v)
            {
            case 0:
                void;
            }
            ext;
        };
    """

    def __init__(
        self,
        hash: Hash,
        header: LedgerHeader,
        ext: LedgerHeaderHistoryEntryExt,
    ) -> None:
        self.hash = hash
        self.header = header
        self.ext = ext

    def pack(self, packer: Packer) -> None:
        self.hash.pack(packer)
        self.header.pack(packer)
        self.ext.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> LedgerHeaderHistoryEntry:
        hash = Hash.unpack(unpacker)
        header = LedgerHeader.unpack(unpacker)
        ext = LedgerHeaderHistoryEntryExt.unpack(unpacker)
        return cls(
            hash=hash,
            header=header,
            ext=ext,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> LedgerHeaderHistoryEntry:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> LedgerHeaderHistoryEntry:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.hash,
                self.header,
                self.ext,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.hash == other.hash
            and self.header == other.header
            and self.ext == other.ext
        )

    def __repr__(self):
        out = [
            f"hash={self.hash}",
            f"header={self.header}",
            f"ext={self.ext}",
        ]
        return f"<LedgerHeaderHistoryEntry [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import Optional

from xdrlib3 import Packer, Unpacker

from .account_id import AccountID
from .signer import Signer
from .string32 import String32
from .uint32 import Uint32

__all__ = ["SetOptionsOp"]


class SetOptionsOp:
    """
    XDR Source Code::

        struct SetOptionsOp
        {
            AccountID* inflationDest; // sets the inflation destination

            uint32* clearFlags; // which flags to clear
            uint32* setFlags;   // which flags to set

            // account threshold manipulation
            uint32* masterWeight; // weight of the master account
            uint32* lowThreshold;
            uint32* medThreshold;
            uint32* highThreshold;

            string32* homeDomain; // sets the home domain

            // Add, update or remove a signer for the account
            // signer is deleted if the weight is 0
            Signer* signer;
        };
    """

    def __init__(
        self,
        inflation_dest: Optional[AccountID],
        clear_flags: Optional[Uint32],
        set_flags: Optional[Uint32],
        master_weight: Optional[Uint32],
        low_threshold: Optional[Uint32],
        med_threshold: Optional[Uint32],
        high_threshold: Optional[Uint32],
        home_domain: Optional[String32],
        signer: Optional[Signer],
    ) -> None:
        self.inflation_dest = inflation_dest
        self.clear_flags = clear_flags
        self.set_flags = set_flags
        self.master_weight = master_weight
        self.low_threshold = low_threshold
        self.med_threshold = med_threshold
        self.high_threshold = high_threshold
        self.home_domain = home_domain
        self.signer = signer

    def pack(self, packer: Packer) -> None:
        if self.inflation_dest is None:
            packer.pack_uint(0)
        else:
            packer.pack_uint(1)
            self.inflation_dest.pack(packer)
        if self.clear_flags is None:
            packer.pack_uint(0)
        else:
            packer.pack_uint(1)
            self.clear_flags.pack(packer)
        if self.set_flags is None:
            packer.pack_uint(0)
        else:
            packer.pack_uint(1)
            self.set_flags.pack(packer)
        if self.master_weight is None:
            packer.pack_uint(0)
        else:
            packer.pack_uint(1)
            self.master_weight.pack(packer)
        if self.low_threshold is None:
            packer.pack_uint(0)
        else:
            packer.pack_uint(1)
            self.low_threshold.pack(packer)
        if self.med_threshold is None:
            packer.pack_uint(0)
        else:
            packer.pack_uint(1)
            self.med_threshold.pack(packer)
        if self.high_threshold is None:
            packer.pack_uint(0)
        else:
            packer.pack_uint(1)
            self.high_threshold.pack(packer)
        if self.home_domain is None:
            packer.pack_uint(0)
        else:
            packer.pack_uint(1)
            self.home_domain.pack(packer)
        if self.signer is None:
            packer.pack_uint(0)
        else:
            packer.pack_uint(1)
            self.signer.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SetOptionsOp:
        inflation_dest = AccountID.unpack(unpacker) if unpacker.unpack_uint() else None
        clear_flags = Uint32.unpack(unpacker) if unpacker.unpack_uint() else None
        set_flags = Uint32.unpack(unpacker) if unpacker.unpack_uint() else None
        master_weight = Uint32.unpack(unpacker) if unpacker.unpack_uint() else None
        low_threshold = Uint32.unpack(unpacker) if unpacker.unpack_uint() else None
        med_threshold = Uint32.unpack(unpacker) if unpacker.unpack_uint() else None
        high_threshold = Uint32.unpack(unpacker) if unpacker.unpack_uint() else None
        home_domain = String32.unpack(unpacker) if unpacker.unpack_uint() else None
        signer = Signer.unpack(unpacker) if unpacker.unpack_uint() else None
        return cls(
            inflation_dest=inflation_dest,
            clear_flags=clear_flags,
            set_flags=set_flags,
            master_weight=master_weight,
            low_threshold=low_threshold,
            med_threshold=med_threshold,
            high_threshold=high_threshold,
            home_domain=home_domain,
            signer=signer,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SetOptionsOp:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SetOptionsOp:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.inflation_dest,
                self.clear_flags,
                self.set_flags,
                self.master_weight,
                self.low_threshold,
                self.med_threshold,
                self.high_threshold,
                self.home_domain,
                self.signer,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.inflation_dest == other.inflation_dest
            and self.clear_flags == other.clear_flags
            and self.set_flags == other.set_flags
            and self.master_weight == other.master_weight
            and self.low_threshold == other.low_threshold
            and self.med_threshold == other.med_threshold
            and self.high_threshold == other.high_threshold
            and self.home_domain == other.home_domain
            and self.signer == other.signer
        )

    def __repr__(self):
        out = [
            f"inflation_dest={self.inflation_dest}",
            f"clear_flags={self.clear_flags}",
            f"set_flags={self.set_flags}",
            f"master_weight={self.master_weight}",
            f"low_threshold={self.low_threshold}",
            f"med_threshold={self.med_threshold}",
            f"high_threshold={self.high_threshold}",
            f"home_domain={self.home_domain}",
            f"signer={self.signer}",
        ]
        return f"<SetOptionsOp [{', '.join(out)}]>"



from __future__ import annotations

import json
from typing import TYPE_CHECKING, Type

from . import Keypair
from . import xdr as stellar_xdr
from .account import Account
from .address import Address
from .base_soroban_server import (
    Durability,
    ResourceLeeway,
    _assemble_transaction,
    _generate_unique_request_id,
)
from .client.aiohttp_client import AiohttpClient
from .exceptions import (
    AccountNotFoundException,
    PrepareTransactionException,
    SorobanRpcErrorResponse,
)
from .soroban_rpc import *

if TYPE_CHECKING:
    from .client.base_async_client import BaseAsyncClient
    from .fee_bump_transaction_envelope import FeeBumpTransactionEnvelope
    from .transaction_envelope import TransactionEnvelope

__all__ = ["SorobanServerAsync", "Durability"]

V = TypeVar("V")


class SorobanServerAsync:
    """Server handles the network connection to a Soroban RPC instance and
    exposes an interface for requests to that instance.

    :param server_url: Soroban RPC server URL. (ex. ``https://soroban-testnet.stellar.org:443``)
    :param client: A client instance that will be used to make requests.
    """

    def __init__(
        self,
        server_url: str = "https://soroban-testnet.stellar.org:443",
        client: Optional[BaseAsyncClient] = None,
    ) -> None:
        self.server_url: str = server_url

        if not client:
            client = AiohttpClient()
        self._client: BaseAsyncClient = client

    async def get_health(self) -> GetHealthResponse:
        """General node health check.

        See `Soroban RPC Documentation - getHealth <https://developers.stellar.org/docs/data/rpc/api-reference/methods/getHealth>`_

        :return: A :class:`GetHealthResponse <stellar_sdk.soroban_rpc.GetHealthResponse>` object.
        :raises: :exc:`SorobanRpcErrorResponse <stellar_sdk.exceptions.SorobanRpcErrorResponse>` - If the Soroban-RPC instance returns an error response.
        """
        request: Request = Request(
            id=_generate_unique_request_id(),
            method="getHealth",
            params=None,
        )
        return await self._post(request, GetHealthResponse)

    async def get_events(
        self,
        start_ledger: int,
        filters: Sequence[EventFilter] = None,
        cursor: str = None,
        limit: int = None,
    ) -> GetEventsResponse:
        """Fetch a list of events that occurred in the ledger range.

        See `Soroban RPC Documentation - getEvents <https://developers.stellar.org/docs/data/rpc/api-reference/methods/getEvents>`_

        :param start_ledger: The first ledger to include in the results.
        :param filters: A list of filters to apply to the results.
        :param cursor: A cursor value for use in pagination.
        :param limit: The maximum number of records to return.
        :return: A :class:`GetEventsResponse <stellar_sdk.soroban_rpc.GetEventsResponse>` object.
        :raises: :exc:`SorobanRpcErrorResponse <stellar_sdk.exceptions.SorobanRpcErrorResponse>` - If the Soroban-RPC instance returns an error response.
        """
        pagination = PaginationOptions(cursor=cursor, limit=limit)
        data = GetEventsRequest(
            startLedger=str(start_ledger),
            filters=filters,
            pagination=pagination,
        )
        request: Request = Request[GetEventsRequest](
            id=_generate_unique_request_id(), method="getEvents", params=data
        )
        return await self._post(request, GetEventsResponse)

    async def get_network(self) -> GetNetworkResponse:
        """General info about the currently configured network.

        See `Soroban RPC Documentation - getNetwork <https://developers.stellar.org/docs/data/rpc/api-reference/methods/getNetwork>`_

        :return: A :class:`GetNetworkResponse <stellar_sdk.soroban_rpc.GetNetworkResponse>` object.
        :raises: :exc:`SorobanRpcErrorResponse <stellar_sdk.exceptions.SorobanRpcErrorResponse>` - If the Soroban-RPC instance returns an error response.
        """
        request: Request = Request(
            id=_generate_unique_request_id(),
            method="getNetwork",
            params=None,
        )
        return await self._post(request, GetNetworkResponse)

    async def get_latest_ledger(self) -> GetLatestLedgerResponse:
        """Fetches the latest ledger meta info from network which Soroban-RPC is connected to.

        See `Soroban RPC Documentation - getLatestLedger <https://developers.stellar.org/docs/data/rpc/api-reference/methods/getLatestLedger>`_

        :return: A :class:`GetLatestLedgerResponse <stellar_sdk.soroban_rpc.GetLatestLedgerResponse>` object.
        :raises: :exc:`SorobanRpcErrorResponse <stellar_sdk.exceptions.SorobanRpcErrorResponse>` - If the Soroban-RPC instance returns an error response.
        """
        request: Request = Request(
            id=_generate_unique_request_id(),
            method="getLatestLedger",
            params=None,
        )
        return await self._post(request, GetLatestLedgerResponse)

    async def get_ledger_entries(
        self, keys: List[stellar_xdr.LedgerKey]
    ) -> GetLedgerEntriesResponse:
        """For reading the current value of ledger entries directly.

        Allows you to directly inspect the current state of a contract, a contract's code,
        or any other ledger entry. This is a backup way to access your contract data
        which may not be available via events or simulateTransaction.

        See `Soroban RPC Documentation - getLedgerEntries <https://developers.stellar.org/docs/data/rpc/api-reference/methods/getLedgerEntries>`_

        :param keys: The ledger keys to fetch.
        :return: A :class:`GetLedgerEntriesResponse <stellar_sdk.soroban_rpc.GetLedgerEntriesResponse>` object.
        :raises: :exc:`SorobanRpcErrorResponse <stellar_sdk.exceptions.SorobanRpcErrorResponse>` - If the Soroban-RPC instance returns an error response.
        """
        request = Request[GetLedgerEntriesRequest](
            id=_generate_unique_request_id(),
            method="getLedgerEntries",
            params=GetLedgerEntriesRequest(keys=[key.to_xdr() for key in keys]),
        )
        return await self._post(request, GetLedgerEntriesResponse)

    async def get_transaction(self, transaction_hash: str) -> GetTransactionResponse:
        """Fetch the specified transaction.

        See `Soroban RPC Documentation - getTransaction <https://developers.stellar.org/docs/data/rpc/api-reference/methods/getTransaction>`_

        :param transaction_hash: The hash of the transaction to fetch.
        :return: A :class:`GetTransactionResponse <stellar_sdk.soroban_rpc.GetTransactionResponse>` object.
        :raises: :exc:`SorobanRpcErrorResponse <stellar_sdk.exceptions.SorobanRpcErrorResponse>` - If the Soroban-RPC instance returns an error response.
        """
        request = Request[GetTransactionRequest](
            id=_generate_unique_request_id(),
            method="getTransaction",
            params=GetTransactionRequest(hash=transaction_hash),
        )
        return await self._post(request, GetTransactionResponse)

    async def simulate_transaction(
        self,
        transaction_envelope: TransactionEnvelope,
        addl_resources: Optional[ResourceLeeway] = None,
    ) -> SimulateTransactionResponse:
        """Submit a trial contract invocation to get back return values, expected ledger footprint, and expected costs.

        See `Soroban RPC Documentation - simulateTransaction <https://developers.stellar.org/docs/data/rpc/api-reference/methods/simulateTransaction>`_

        :param transaction_envelope: The transaction to simulate. It should include exactly one operation,
            which must be one of :class:`RestoreFootprint <stellar_sdk.operation.RestoreFootprintOperation>`,
            :class:`InvokeHostFunction <stellar_sdk.operation.InvokeHostFunction>` or
            :class:`ExtendFootprintTTL <stellar_sdk.operation.RestoreFootprint>` operation.
            Any provided footprint will be ignored.
        :param addl_resources: Additional resource include in the simulation.
        :return: A :class:`SimulateTransactionResponse <stellar_sdk.soroban_rpc.SimulateTransactionResponse>` object
            contains the cost, footprint, result/auth requirements (if applicable), and error of the transaction.
        """
        xdr = (
            transaction_envelope
            if isinstance(transaction_envelope, str)
            else transaction_envelope.to_xdr()
        )
        resource_config = None
        if addl_resources:
            resource_config = ResourceConfig(
                instruction_lee_way=addl_resources.cpu_instructions,
            )
        request = Request[SimulateTransactionRequest](
            id=_generate_unique_request_id(),
            method="simulateTransaction",
            params=SimulateTransactionRequest(
                transaction=xdr, resource_config=resource_config
            ),
        )
        return await self._post(request, SimulateTransactionResponse)

    async def send_transaction(
        self,
        transaction_envelope: Union[
            TransactionEnvelope, FeeBumpTransactionEnvelope, str
        ],
    ) -> SendTransactionResponse:
        """Submit a real transaction to the Stellar network. This is the only way to make changes "on-chain".

        See `Soroban RPC Documentation - sendTransaction <https://developers.stellar.org/docs/data/rpc/api-reference/methods/sendTransaction>`_

        :param transaction_envelope: The transaction to send.
        :return: A :class:`SendTransactionResponse <stellar_sdk.soroban_rpc.SendTransactionResponse>` object.
        :raises: :exc:`SorobanRpcErrorResponse <stellar_sdk.exceptions.SorobanRpcErrorResponse>` - If the Soroban-RPC instance returns an error response.
        """
        xdr = (
            transaction_envelope
            if isinstance(transaction_envelope, str)
            else transaction_envelope.to_xdr()
        )
        request = Request[SendTransactionRequest](
            id=_generate_unique_request_id(),
            method="sendTransaction",
            params=SendTransactionRequest(transaction=xdr),
        )
        return await self._post(request, SendTransactionResponse)

    async def get_fee_stats(self) -> GetFeeStatsResponse:
        """General info about the fee stats.

        See `Soroban RPC Documentation - getFeeStats <https://developers.stellar.org/docs/data/rpc/api-reference/methods/getFeeStats>`_

        :return: A :class:`GetFeeStatsResponse <stellar_sdk.soroban_rpc.GetFeeStatsResponse>` object.
        :raises: :exc:`SorobanRpcErrorResponse <stellar_sdk.exceptions.SorobanRpcErrorResponse>` - If the Soroban-RPC instance returns an error response.
        """
        request: Request = Request(
            id=_generate_unique_request_id(),
            method="getFeeStats",
            params=None,
        )
        return await self._post(request, GetFeeStatsResponse)

    async def get_transactions(
        self,
        start_ledger: int,
        cursor: str = None,
        limit: int = None,
    ) -> GetTransactionsResponse:
        """Fetch a detailed list of transactions starting from the user specified starting point that you can paginate
        as long as the pages fall within the history retention of their corresponding RPC provider.

        See `Soroban RPC Documentation - getTransactions <https://developers.stellar.org/docs/data/rpc/api-reference/methods/getTransactions>`_

        :param start_ledger: The first ledger to include in the results.
        :param cursor: A cursor value for use in pagination.
        :param limit: The maximum number of records to return.
        :return: A :class:`GetTransactionsResponse <stellar_sdk.soroban_rpc.GetTransactionsResponse>` object.
        :raises: :exc:`SorobanRpcErrorResponse <stellar_sdk.exceptions.SorobanRpcErrorResponse>` - If the Soroban-RPC instance returns an error response.
        """
        pagination = PaginationOptions(cursor=cursor, limit=limit)
        data = GetTransactionsRequest(
            startLedger=str(start_ledger),
            pagination=pagination,
        )
        request: Request = Request[GetTransactionsRequest](
            id=_generate_unique_request_id(), method="getTransactions", params=data
        )
        return await self._post(request, GetTransactionsResponse)

    async def get_ledgers(
        self, start_ledger: int, cursor: str = None, limit: int = None
    ) -> GetLedgersResponse:
        """Fetch a detailed list of ledgers starting from the user specified starting point that you can paginate
        as long as the pages fall within the history retention of their corresponding RPC provider.

        See `Soroban RPC Documentation - getLedgers <https://developers.stellar.org/docs/data/rpc/api-reference/methods/getLedgers>`_

        :param start_ledger: The first ledger to include in the results.
        :param cursor: A cursor value for use in pagination.
        :param limit: The maximum number of records to return.
        :return: A :class:`GetLedgersResponse <stellar_sdk.soroban_rpc.GetLedgersResponse>` object.
        :raises: :exc:`SorobanRpcErrorResponse <stellar_sdk.exceptions.SorobanRpcErrorResponse>` - If the Soroban-RPC instance returns an error response.
        """
        pagination = PaginationOptions(cursor=cursor, limit=limit)
        data = GetLedgersRequest(
            startLedger=str(start_ledger),
            pagination=pagination,
        )
        request: Request = Request[GetLedgersRequest](
            id=_generate_unique_request_id(), method="getLedgers", params=data
        )
        return await self._post(request, GetLedgersResponse)

    async def load_account(self, account_id: str) -> Account:
        """Load an account from the server, you can use the returned account
        object as source account for transactions.

        :param account_id: The account ID.
        :return: An :class:`Account <stellar_sdk.account.Account>` object.
        :raises: :exc:`AccountNotFoundException <stellar_sdk.exceptions.AccountNotFoundException>` - If the account is not found on the network.
        :raises: :exc:`SorobanRpcErrorResponse <stellar_sdk.exceptions.SorobanRpcErrorResponse>` - If the Soroban-RPC instance returns an error response.
        """
        account_id_xdr = Keypair.from_public_key(account_id).xdr_account_id()
        key = stellar_xdr.LedgerKey(
            stellar_xdr.LedgerEntryType.ACCOUNT,
            account=stellar_xdr.LedgerKeyAccount(account_id=account_id_xdr),
        )

        resp = await self.get_ledger_entries([key])
        if not resp.entries:
            raise AccountNotFoundException(account_id)
        assert len(resp.entries) == 1
        data = stellar_xdr.LedgerEntryData.from_xdr(resp.entries[0].xdr)
        assert data.account is not None
        return Account(account_id, data.account.seq_num.sequence_number.int64)

    async def get_contract_data(
        self,
        contract_id: str,
        key: stellar_xdr.SCVal,
        durability: Durability = Durability.PERSISTENT,
    ) -> Optional[LedgerEntryResult]:
        """Reads the current value of contract data ledger entries directly.

        :param contract_id: The contract ID containing the data to load. Encoded as Stellar Contract Address,
            for example: ``"CCJZ5DGASBWQXR5MPFCJXMBI333XE5U3FSJTNQU7RIKE3P5GN2K2WYD5"``
        :param key: The key of the contract data to load.
        :param durability: The "durability keyspace" that this ledger key belongs to, which is either
            :class:`Durability.TEMPORARY` or :class:`Durability.PERSISTENT`. Defaults to :class:`Durability.PERSISTENT`.
        :return: A :class:`LedgerEntryResult <stellar_sdk.soroban_rpc.LedgerEntryResult>` object contains the ledger entry result or ``None`` if not found.
        :raises: :exc:`SorobanRpcErrorResponse <stellar_sdk.exceptions.SorobanRpcErrorResponse>` - If the Soroban-RPC instance returns an error response.
        """
        sc_address = Address(contract_id).to_xdr_sc_address()
        xdr_durability = (
            stellar_xdr.ContractDataDurability.PERSISTENT
            if durability == Durability.PERSISTENT
            else stellar_xdr.ContractDataDurability.TEMPORARY
        )
        contract_key = stellar_xdr.LedgerKey(
            stellar_xdr.LedgerEntryType.CONTRACT_DATA,
            contract_data=stellar_xdr.LedgerKeyContractData(
                contract=sc_address,
                key=key,
                durability=xdr_durability,
            ),
        )
        resp = await self.get_ledger_entries([contract_key])
        entries = resp.entries
        if not entries:
            return None
        return entries[0]

    async def get_version_info(self) -> GetVersionInfoResponse:
        """Version information about the RPC and Captive core.

        See `Soroban RPC Documentation - getVersionInfo <https://developers.stellar.org/docs/data/rpc/api-reference/methods/getVersionInfo>`_

        :return: A :class:`GetVersionInfoResponse <stellar_sdk.soroban_rpc.GetVersionInfoResponse>` object.
        :raises: :exc:`SorobanRpcErrorResponse <stellar_sdk.exceptions.SorobanRpcErrorResponse>` - If the Soroban-RPC instance returns an error response.
        """
        request: Request = Request(
            id=_generate_unique_request_id(),
            method="getVersionInfo",
            params=None,
        )
        return await self._post(request, GetVersionInfoResponse)

    async def prepare_transaction(
        self,
        transaction_envelope: TransactionEnvelope,
        simulate_transaction_response: SimulateTransactionResponse = None,
    ) -> TransactionEnvelope:
        """Submit a trial contract invocation, first run a simulation of the contract
        invocation as defined on the incoming transaction, and apply the results to
        a new copy of the transaction which is then returned. Setting the ledger
        footprint and authorization, so the resulting transaction is ready for signing
        and sending.

        The returned transaction will also have an updated fee that is the sum of fee
        set on incoming transaction with the contract resource fees estimated from
        simulation. It is advisable to check the fee on returned transaction and validate
        or take appropriate measures for interaction with user to confirm it is acceptable.

        You can call the :meth:`simulate_transaction` method directly first if you
        want to inspect estimated fees for a given transaction in detail first if that is
        of importance.

        :param transaction_envelope: The transaction to prepare. It should include exactly one operation, which
            must be one of :py:class:`RestoreFootprint <stellar_sdk.operation.RestoreFootprint>`,
            :py:class:`ExtendFootprintTTL <stellar_sdk.operation.ExtendFootprintTTL>`,
            or :py:class:`InvokeHostFunction <stellar_sdk.operation.InvokeHostFunction>`. Any provided
            footprint will be ignored. You can use :meth:`stellar_sdk.Transaction.is_soroban_transaction` to check
            if a transaction is a Soroban transaction. Any provided footprint will be overwritten.
            However, if your operation has existing auth entries, they will be preferred over ALL auth
            entries from the simulation. In other words, if you include auth entries, you don't care
            about the auth returned from the simulation. Other fields (footprint, etc.) will be filled
            as normal.
        :param simulate_transaction_response: The response of the simulation of the transaction,
            typically you don't need to pass this parameter, it will be automatically called if you don't pass it.
        :return: A copy of the :class:`TransactionEnvelope <stellar_sdk.transaction_envelope.TransactionEnvelope>`,
            with the expected authorizations (in the case of invocation) and ledger footprint added.
            The transaction fee will also automatically be padded with the contract's minimum resource fees
            discovered from the simulation.
        """
        if not simulate_transaction_response:
            simulate_transaction_response = await self.simulate_transaction(
                transaction_envelope
            )
        if simulate_transaction_response.error:
            raise PrepareTransactionException(
                "Simulation transaction failed, the response contains error information.",
                simulate_transaction_response,
            )
        te = _assemble_transaction(transaction_envelope, simulate_transaction_response)
        return te

    async def _post(self, request_body: Request, response_body_type: Type[V]) -> V:
        json_data = request_body.model_dump_json(by_alias=True)
        data = await self._client.post(
            self.server_url,
            json_data=json.loads(json_data),
        )
        response = Response[response_body_type].model_validate(data.json())  # type: ignore[valid-type]
        if response.error:
            raise SorobanRpcErrorResponse(
                response.error.code, response.error.message, response.error.data
            )
        assert response.result is not None
        return response.result

    async def close(self) -> None:
        """Close underlying connector, and release all acquired resources."""
        await self._client.close()

    async def __aenter__(self) -> "SorobanServerAsync":
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.close()

    def __repr__(self):
        return f"<SorobanServerAsync [server_url={self.server_url}, client={self._client}]>"



import binascii
import math
import os
import time
import warnings
from decimal import Decimal
from typing import List, Optional, Sequence, Union

from . import StrKey
from . import xdr as stellar_xdr
from .account import Account
from .address import Address
from .asset import Asset
from .fee_bump_transaction import FeeBumpTransaction
from .fee_bump_transaction_envelope import FeeBumpTransactionEnvelope
from .keypair import Keypair
from .ledger_bounds import LedgerBounds
from .liquidity_pool_asset import LiquidityPoolAsset
from .liquidity_pool_id import LiquidityPoolId
from .memo import *
from .muxed_account import MuxedAccount
from .network import Network
from .operation import *
from .preconditions import Preconditions
from .price import Price
from .signer import Signer
from .signer_key import SignedPayloadSigner, SignerKey
from .soroban_data_builder import SorobanDataBuilder
from .time_bounds import TimeBounds
from .transaction import Transaction
from .transaction_envelope import TransactionEnvelope
from .utils import hex_to_bytes, is_valid_hash

__all__ = ["TransactionBuilder"]

MIN_BASE_FEE = 100


class TransactionBuilder:
    """Transaction builder helps constructs a new :class:`TransactionEnvelope
    <stellar_sdk.transaction_envelope.TransactionEnvelope>` using the given
    :class:`Account <stellar_sdk.account.Account>` as the transaction's "source account". The transaction will use
    the current sequence number of the given account as its sequence number and increment the given account's
    sequence number by one.

    Operations can be added to the transaction via their corresponding builder
    methods, and each returns the :class:`TransactionEnvelope <stellar_sdk.transaction_envelope.TransactionEnvelope>`
    object, so they can be chained together. After adding the desired operations, call
    the :func:`build` method on the TransactionBuilder to return a fully constructed
    :class:`TransactionEnvelope <stellar_sdk.transaction_envelope.TransactionEnvelope>` that can be signed.

    Be careful about **unsubmitted transactions**! When you build a transaction, stellar-sdk automatically
    increments the source account's sequence number. If you end up not submitting this transaction and submitting
    another one instead, it'll fail due to the sequence number being wrong. So if you decide not to use a built
    transaction, make sure to update the source account's sequence number
    with :func:`stellar_sdk.server.Server.load_account` or :func:`stellar_sdk.server_async.ServerAsync.load_account`
    before creating another transaction.

    The following code example creates a new transaction with :class:`CreateAccount <stellar_sdk.operation.CreateAccount>`
    and :class:`Payment <stellar_sdk.operation.Payment>` operations. The Transaction's source account(alice) first
    funds `bob`, then sends a payment to `bob`. The built transaction is then signed by `alice_keypair`::

        # Alice funds Bob with 5 XLM and then pays Bob 10.25 XLM
        from stellar_sdk import Server, Asset, Keypair, TransactionBuilder, Network

        alice_keypair = Keypair.from_secret("SBFZCHU5645DOKRWYBXVOXY2ELGJKFRX6VGGPRYUWHQ7PMXXJNDZFMKD")
        bob_address = "GA7YNBW5CBTJZ3ZZOWX3ZNBKD6OE7A7IHUQVWMY62W2ZBG2SGZVOOPVH"

        server = Server("https://horizon-testnet.stellar.org")
        alice_account = server.load_account(alice_keypair.public_key)
        network_passphrase = Network.TESTNET_NETWORK_PASSPHRASE
        base_fee = 100
        transaction = (
            TransactionBuilder(
                source_account=alice_account,
                network_passphrase=network_passphrase,
                base_fee=base_fee,
            )
                .add_text_memo("Hello, Stellar!")
                .append_create_account_op(bob_address, "5")
                .append_payment_op(bob_address, Asset.native(), "10.25")
                .set_timeout(30)
                .build()
        )
        transaction.sign(alice_keypair)
        response = server.submit_transaction(transaction)
        print(response)

    :param source_account: The source account for this transaction.
    :param network_passphrase: The network to connect to for verifying and retrieving additional attributes from.
        Defaults to ``Test SDF Network ; September 2015``.
    :param base_fee: Max fee you're willing to pay per operation in this transaction (**in stroops**).
    :param v1: When this value is set to True, V1 transactions will be generated,
        otherwise V0 transactions will be generated.
        See `CAP-0015 <https://github.com/stellar/stellar-protocol/blob/master/core/cap-0015.md>`_ for more information.
    """

    def __init__(
        self,
        source_account: Account,
        network_passphrase: str = Network.TESTNET_NETWORK_PASSPHRASE,
        base_fee: int = MIN_BASE_FEE,
        v1: bool = True,
    ):
        self.source_account: Account = source_account
        self.base_fee: int = base_fee
        self.network_passphrase: str = network_passphrase
        self.operations: List[Operation] = []
        self.time_bounds: Optional[TimeBounds] = None
        self.ledger_bounds: Optional[LedgerBounds] = None
        self.min_sequence_number: Optional[int] = None
        self.min_sequence_age: Optional[int] = None
        self.min_sequence_ledger_gap: Optional[int] = None
        self.extra_signers: List[SignerKey] = []
        self.memo: Memo = NoneMemo()
        self.v1: bool = v1

        self.soroban_data: Optional[stellar_xdr.SorobanTransactionData] = None

    def build(self) -> TransactionEnvelope:
        """This will build the transaction envelope.
        It will also increment the source account's sequence number by 1.

        :return: New transaction envelope.
        """
        if self.time_bounds is None:
            warnings.warn(
                "It looks like you haven't set a TimeBounds for the transaction, "
                "we strongly recommend that you set it. "
                "You can learn why you should set it up through this link: "
                "https://www.stellar.org/developers-blog/transaction-submission-timeouts-and-dynamic-fees-faq"
            )

        source = self.source_account.account
        sequence = self.source_account.sequence + 1
        preconditions = Preconditions(
            time_bounds=self.time_bounds,
            ledger_bounds=self.ledger_bounds,
            min_sequence_number=self.min_sequence_number,
            min_sequence_age=self.min_sequence_age,
            min_sequence_ledger_gap=self.min_sequence_ledger_gap,
            extra_signers=self.extra_signers,
        )
        fee = self.base_fee * len(self.operations)
        if self.soroban_data:
            fee += self.soroban_data.resource_fee.int64
        transaction = Transaction(
            source=source,
            sequence=sequence,
            fee=fee,
            operations=self.operations,
            memo=self.memo,
            preconditions=preconditions,
            soroban_data=self.soroban_data,
            v1=self.v1,
        )
        transaction_envelope = TransactionEnvelope(
            transaction=transaction, network_passphrase=self.network_passphrase
        )
        self.source_account.increment_sequence_number()
        return transaction_envelope

    @staticmethod
    def build_fee_bump_transaction(
        fee_source: Union[MuxedAccount, Keypair, str],
        base_fee: int,
        inner_transaction_envelope: TransactionEnvelope,
        network_passphrase: str = Network.TESTNET_NETWORK_PASSPHRASE,
    ) -> FeeBumpTransactionEnvelope:
        """Create a
        :py:class:`FeeBumpTransactionEnvelope <stellar_sdk.fee_bump_transaction_envelope.FeeBumpTransactionEnvelope>`
        object.

        See `CAP-0015 <https://github.com/stellar/stellar-protocol/blob/master/core/cap-0015.md>`_ for more information.

        :param fee_source: The account paying for the transaction.
        :param base_fee: The max fee willing to pay per operation in inner transaction (**in stroops**).
        :param inner_transaction_envelope: The TransactionEnvelope to be bumped by the fee bump transaction.
        :param network_passphrase: The network to connect to for verifying and retrieving additional attributes from.
        :return: a :class:`TransactionBuilder <stellar_sdk.transaction_envelope.TransactionBuilder>` via the XDR object.
        """

        if base_fee < MIN_BASE_FEE:
            raise ValueError(
                f"Invalid `base_fee`, it should be at least {MIN_BASE_FEE} stroops."
            )

        soroban_resource_fee = 0
        if inner_transaction_envelope.transaction.soroban_data:
            soroban_resource_fee = (
                inner_transaction_envelope.transaction.soroban_data.resource_fee.int64
            )

        inner_include_fee = (
            inner_transaction_envelope.transaction.fee - soroban_resource_fee
        )  # dont include soroban resource fee
        inner_base_fee = math.ceil(
            inner_include_fee / len(inner_transaction_envelope.transaction.operations)
        )

        if base_fee < inner_base_fee:
            raise ValueError(
                f"Invalid `base_fee`, it should be at least {inner_base_fee} stroops."
            )

        fee = base_fee * (len(inner_transaction_envelope.transaction.operations) + 1)
        fee += soroban_resource_fee

        fee_bump_transaction = FeeBumpTransaction(
            fee_source=fee_source,
            fee=fee,
            inner_transaction_envelope=inner_transaction_envelope,
        )
        transaction_envelope = FeeBumpTransactionEnvelope(
            transaction=fee_bump_transaction,
            network_passphrase=network_passphrase,
        )
        return transaction_envelope

    @staticmethod
    def from_xdr(
        xdr: str, network_passphrase: str
    ) -> Union[TransactionEnvelope, FeeBumpTransactionEnvelope]:
        """When you are not sure whether your XDR belongs to
        :py:class:`TransactionEnvelope <stellar_sdk.transaction_envelope.TransactionEnvelope>`
        or :py:class:`FeeBumpTransactionEnvelope <stellar_sdk.fee_bump_transaction_envelope.FeeBumpTransactionEnvelope>`,
        you can use this function.

        An example::

            from stellar_sdk import Network, TransactionBuilder

            xdr = "AAAAAgAAAADHJNEDn33/C1uDkDfzDfKVq/4XE9IxDfGiLCfoV7riZQAAA+gCI4TVABpRPgAAAAAAAAAAAAAAAQAAAAAAAAADAAAAAUxpcmEAAAAAabIaDgm0ypyJpsVfEjZw2mO3Enq4Q4t5URKfWtqukSUAAAABVVNEAAAAAADophqGHmCvYPgHc+BjRuXHLL5Z3K3aN2CNWO9CUR2f3AAAAAAAAAAAE8G9mAADcH8AAAAAMYdBWgAAAAAAAAABV7riZQAAAEARGCGwYk/kEB2Z4UL20y536evnwmmSc4c2FnxlvUcPZl5jgWHcNwY8LTpFhdrUN9TZWciCRp/JCZYa0SJh8cYB"
            te = TransactionBuilder.from_xdr(xdr, Network.PUBLIC_NETWORK_PASSPHRASE)
            print(te)

        :param xdr: Transaction envelope XDR
        :param network_passphrase: The network to connect to for verifying and retrieving
            additional attributes from. (ex. ``"Public Global Stellar Network ; September 2015"``)
        :raises: :exc:`ValueError <stellar_sdk.exceptions.ValueError>` - XDR is neither :py:class:`TransactionEnvelope <stellar_sdk.transaction_envelope.TransactionEnvelope>`
            nor :py:class:`FeeBumpTransactionEnvelope <stellar_sdk.fee_bump_transaction_envelope.FeeBumpTransactionEnvelope>`
        """
        if FeeBumpTransactionEnvelope.is_fee_bump_transaction_envelope(xdr):
            return FeeBumpTransactionEnvelope.from_xdr(xdr, network_passphrase)
        return TransactionEnvelope.from_xdr(xdr, network_passphrase)

    def add_time_bounds(self, min_time: int, max_time: int) -> "TransactionBuilder":
        """Sets a timeout precondition on the transaction.

        Because of the distributed nature of the Stellar network it is possible
        that the status of your transaction will be determined after a long time
        if the network is highly congested. If you want to be sure to receive the
        status of the transaction within a given period you should set
        the :class:`TimeBounds` with `max_time` on the transaction (this is
        what :func:`set_timeout` does internally).

        Please note that Horizon may still return **504 Gateway Timeout**
        error, even for short timeouts. In such case you need to resubmit the same
        transaction again without making any changes to receive a status. This
        method is using the machine system time (UTC), make sure it is set
        correctly.

        Add a UNIX timestamp, determined by ledger time, of a lower and
        upper bound of when this transaction will be valid. If a transaction is
        submitted too early or too late, it will fail to make it into the
        transaction set. `max_time` equal ``0`` means that it's not set.

        :param min_time: the UNIX timestamp (in seconds)
        :param max_time: the UNIX timestamp (in seconds)
        :return: This builder instance.

        """
        self.time_bounds = TimeBounds(min_time, max_time)
        return self

    def set_timeout(self, timeout: int) -> "TransactionBuilder":
        """Set timeout for the transaction, actually set a :class:`TimeBounds`.

        :param timeout: timeout in second.
        :return: This builder instance.
        :raises:
            :exc:`ValueError <stellar_sdk.exceptions.ValueError>`: if `time_bound` is already set.
        """
        if self.time_bounds:
            raise ValueError(
                "TimeBounds has been already set - setting timeout would overwrite it."
            )
        timeout_timestamp = int(time.time()) + timeout
        self.time_bounds = TimeBounds(min_time=0, max_time=timeout_timestamp)
        return self

    def set_ledger_bounds(
        self, min_ledger: int, max_ledger: int
    ) -> "TransactionBuilder":
        """If you want to prepare a transaction which will only
        be valid within some range of ledgers, you can set a `ledger_bounds` precondition.
        Internally this will set the :class:`LedgerBounds` preconditions.

        :param min_ledger: The minimum ledger this transaction is valid at, or after.
            Cannot be negative. If the value is ``0``, the transaction is valid immediately.
        :param max_ledger: The maximum ledger this transaction is valid before.
            Cannot be negative. If the value is ``0``, the transaction is valid indefinitely.
        :return: This builder instance.
        """
        self.ledger_bounds = LedgerBounds(min_ledger, max_ledger)
        return self

    def set_min_sequence_number(self, min_sequence_number: int) -> "TransactionBuilder":
        """If you want to prepare a transaction which will be valid only while the account sequence number is
        **min_sequence_number <= source_account_sequence_number < tx.sequence**.

        Note that after execution the account's sequence number is always raised to `tx.sequence`.
        Internally this will set the `min_sequence_number` precondition.

        :param min_sequence_number: The minimum source account sequence
            number this transaction is valid for. If the value is ``None``
            the transaction is valid when **source account's sequence number == tx.sequence - 1**.
        :return: This builder instance.
        """
        self.min_sequence_number = min_sequence_number
        return self

    def set_min_sequence_age(self, min_sequence_age: int) -> "TransactionBuilder":
        """For the transaction to be valid, the current ledger time must be
        at least `min_sequence_age` greater than source account's `sequence_time`.
        Internally this will set the `min_sequence_age` precondition.

        :param min_sequence_age: The minimum amount of time between
            source account sequence time and the ledger time when this transaction
            will become valid. If the value is ``0`` or ``None``, the transaction is unrestricted
            by the account sequence age. Cannot be negative.
        :return: This builder instance.
        """
        self.min_sequence_age = min_sequence_age
        return self

    def set_min_sequence_ledger_gap(
        self, min_sequence_ledger_gap: int
    ) -> "TransactionBuilder":
        """For the transaction to be valid, the current ledger number must be at least
        `min_sequence_ledger_gap` greater than source account's ledger sequence.
        Internally this will set the `min_sequence_ledger_gap` precondition.

        :param min_sequence_ledger_gap: The minimum number of ledgers between source account
            sequence and the ledger number when this transaction will become valid.
            If the value is ``0`` or ``None``, the transaction is unrestricted by the account sequence
            ledger. Cannot be negative.
        :return: This builder instance.
        """
        self.min_sequence_ledger_gap = min_sequence_ledger_gap
        return self

    def set_soroban_data(
        self, soroban_data: Union[stellar_xdr.SorobanTransactionData, str]
    ) -> "TransactionBuilder":
        """Set the SorobanTransactionData. For non-contract(non-Soroban) transactions, this setting has no effect.

        In the case of Soroban transactions, set to an instance of
        SorobanTransactionData. This can typically be obtained from the simulation
        response based on a transaction with a InvokeHostFunctionOp.
        It provides necessary resource estimations for contract invocation.

        :param soroban_data: The SorobanTransactionData as XDR object or base64 encoded string.
        :return: This builder instance.
        """
        self.soroban_data = SorobanDataBuilder.from_xdr(soroban_data).build()
        return self

    def add_extra_signer(
        self, signer_key: Union[SignerKey, SignedPayloadSigner, str]
    ) -> "TransactionBuilder":
        """For the transaction to be valid, there must be a signature corresponding to every
        Signer in this array, even if the signature is not otherwise required by
        the source account or operations.
        Internally this will set the :class:`SignerKey` precondition.

        :param signer_key: The signer key
        :return: This builder instance.
        """
        if isinstance(signer_key, str):
            signer_key = SignerKey.from_encoded_signer_key(signer_key)
        if isinstance(signer_key, SignedPayloadSigner):
            signer_key = SignerKey.ed25519_signed_payload(signer_key)
        self.extra_signers.append(signer_key)
        return self

    def add_memo(self, memo: Memo) -> "TransactionBuilder":
        """Set the memo for the transaction build by this :class:`Builder`.

        :param memo: A memo to add to this transaction.
        :return: This builder instance.

        """
        self.memo = memo
        return self

    def add_text_memo(self, memo_text: Union[str, bytes]) -> "TransactionBuilder":
        """Set the memo for the transaction to a new :class:`TextMemo
        <stellar_sdk.memo.TextMemo>`.

        :param memo_text: The text for the memo to add.
        :return: This builder instance.
        :raises: :exc:`MemoInvalidException <stellar_sdk.exceptions.MemoInvalidException>`:
            if `memo_text` is not a valid text memo.
        """
        memo = TextMemo(memo_text)
        return self.add_memo(memo)

    def add_id_memo(self, memo_id: int) -> "TransactionBuilder":
        """Set the memo for the transaction to a new :class:`IdMemo
        <stellar_sdk.memo.IdMemo>`.

        :param memo_id: A 64 bit unsigned integer to set as the memo.
        :return: This builder instance.
        :raises:
            :exc:`MemoInvalidException <stellar_sdk.exceptions.MemoInvalidException>`:
            if `memo_id` is not a valid id memo.

        """
        memo = IdMemo(memo_id)
        return self.add_memo(memo)

    def add_hash_memo(self, memo_hash: Union[bytes, str]) -> "TransactionBuilder":
        """Set the memo for the transaction to a new :class:`HashMemo
        <stellar_sdk.memo.HashMemo>`.

        :param memo_hash: A 32 byte hash or hex encoded string to use as the memo.
        :return: This builder instance.
        :raises:
            :exc:`MemoInvalidException <stellar_sdk.exceptions.MemoInvalidException>`:
            if `memo_hash` is not a valid hash memo.
        """
        memo = HashMemo(memo_hash)
        return self.add_memo(memo)

    def add_return_hash_memo(
        self, memo_return: Union[bytes, str]
    ) -> "TransactionBuilder":
        """Set the memo for the transaction to a new :class:`RetHashMemo
        <stellar_sdk.memo.ReturnHashMemo>`.

        :param memo_return: A 32 byte hash or hex encoded string intended to be interpreted as
            the hash of the transaction the sender is refunding.
        :return: This builder instance.
        :raises:
            :exc:`MemoInvalidException <stellar_sdk.exceptions.MemoInvalidException>`:
            if `memo_return` is not a valid return hash memo.
        """
        memo = ReturnHashMemo(memo_return)
        return self.add_memo(memo)

    def append_operation(self, operation: Operation) -> "TransactionBuilder":
        """Add an operation to the builder instance

        :param operation: an operation
        :return: This builder instance.
        """
        self.operations.append(operation)
        return self

    def append_create_account_op(
        self,
        destination: str,
        starting_balance: Union[str, Decimal],
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> "TransactionBuilder":
        """Append a :class:`CreateAccount
        <stellar_sdk.operation.CreateAccount>` operation to the list of
        operations.

        :param destination: Account address that is created and funded.
        :param starting_balance: Amount of XLM to send to the newly created
            account. This XLM comes from the source account.
        :param source: The source address to deduct funds from to fund the
            new account.
        :return: This builder instance.

        """
        op = CreateAccount(destination, starting_balance, source)
        return self.append_operation(op)

    def append_change_trust_op(
        self,
        asset: Union[Asset, LiquidityPoolAsset],
        limit: Union[str, Decimal] = None,
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> "TransactionBuilder":
        """Append a :class:`ChangeTrust <stellar_sdk.operation.ChangeTrust>`
        operation to the list of operations.

        :param asset: The asset for the trust line.
        :param limit: The limit for the asset, defaults to max int64(``922337203685.4775807``).
            If the limit is set to ``"0"`` it deletes the trustline.
        :param source: The source address to add the trustline to.
        :return: This builder instance.
        """

        op = ChangeTrust(asset, limit, source)
        return self.append_operation(op)

    def append_payment_op(
        self,
        destination: Union[MuxedAccount, str],
        asset: Asset,
        amount: Union[str, Decimal],
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> "TransactionBuilder":
        """Append a :class:`Payment <stellar_sdk.operation.Payment>` operation
        to the list of operations.

        :param destination: The destination account ID.
        :param asset: The asset to send.
        :param amount: The amount to send.
        :param source: The source account for the operation. Defaults to the
            transaction's source account.
        :return: This builder instance.
        """

        op = Payment(destination, asset, amount, source)
        return self.append_operation(op)

    def append_path_payment_strict_receive_op(
        self,
        destination: Union[MuxedAccount, str],
        send_asset: Asset,
        send_max: Union[str, Decimal],
        dest_asset: Asset,
        dest_amount: Union[str, Decimal],
        path: Sequence[Asset],
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> "TransactionBuilder":
        """Append a :class:`PathPaymentStrictReceive <stellar_sdk.operation.PathPaymentStrictReceive>`
        operation to the list of operations.

        :param destination: The destination account to send to.
        :param send_asset: The `asset` to pay with.
        :param send_max: The maximum amount of `send_asset` to send.
        :param dest_asset: The asset the `destination` will receive.
        :param dest_amount: The amount the `destination` receives.
        :param path: A list of Asset objects to use as the path.
        :param source: The source account for the operation. Defaults to the
            transaction's source account.
        :return: This builder instance.
        """

        op = PathPaymentStrictReceive(
            destination=destination,
            send_asset=send_asset,
            send_max=send_max,
            dest_asset=dest_asset,
            dest_amount=dest_amount,
            path=path,
            source=source,
        )
        return self.append_operation(op)

    def append_path_payment_strict_send_op(
        self,
        destination: Union[MuxedAccount, str],
        send_asset: Asset,
        send_amount: Union[str, Decimal],
        dest_asset: Asset,
        dest_min: Union[str, Decimal],
        path: Sequence[Asset],
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> "TransactionBuilder":
        """Append a :class:`PathPaymentStrictSend <stellar_sdk.operation.PathPaymentStrictSend>`
        operation to the list of operations.

        :param destination: The destination account to send to.
        :param send_asset: The `asset` to pay with.
        :param send_amount: Amount of `send_asset` to send.
        :param dest_asset: The asset the `destination` will receive.
        :param dest_min: The minimum amount of `dest_asset` to be received.
        :param path: A list of Asset objects to use as the path.
        :param source: The source account for the operation. Defaults to the
            transaction's source account.
        :return: This builder instance.
        """
        op = PathPaymentStrictSend(
            destination=destination,
            send_asset=send_asset,
            send_amount=send_amount,
            dest_asset=dest_asset,
            dest_min=dest_min,
            path=path,
            source=source,
        )
        return self.append_operation(op)

    def append_allow_trust_op(
        self,
        trustor: str,
        asset_code: str,
        authorize: Union[TrustLineEntryFlag, bool],
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> "TransactionBuilder":
        """Append an :class:`AllowTrust <stellar_sdk.operation.AllowTrust>`
        operation to the list of operations.

        :param trustor: The account of the recipient of the trustline.
        :param asset_code:  The asset of the trustline the source account
            is authorizing. For example, if an anchor wants to allow another
            account to hold its USD credit, the type is USD:anchor.
        :param authorize: `True` to authorize the line, `False` to deauthorize，if you need further control,
            you can also use :class:`stellar_sdk.operation.allow_trust.TrustLineEntryFlag`.
        :param source: The source address that is establishing the trust in
            the allow trust operation.
        :return: This builder instance.

        """
        warnings.warn(
            "Use `stellar_sdk.transaction_builder.TransactionBuilder.append_set_trust_line_flags_op` instead.",
            DeprecationWarning,
        )
        op = AllowTrust(trustor, asset_code, authorize, source)
        return self.append_operation(op)

    def append_set_options_op(
        self,
        inflation_dest: str = None,
        clear_flags: Union[int, AuthorizationFlag] = None,
        set_flags: Union[int, AuthorizationFlag] = None,
        master_weight: int = None,
        low_threshold: int = None,
        med_threshold: int = None,
        high_threshold: int = None,
        home_domain: str = None,
        signer: Signer = None,
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> "TransactionBuilder":
        """Append a :class:`SetOptions <stellar_sdk.operation.SetOptions>`
        operation to the list of operations.

        :param inflation_dest: Account of the inflation destination.
        :param clear_flags: Indicates which flags to clear. For details about the flags,
            please refer to the `Control Access to an Asset - Flag <https://developers.stellar.org/docs/issuing-assets/control-asset-access/#flags>`__.
            The `bit mask <https://en.wikipedia.org/wiki/Bit_field>`_ integer subtracts from the existing flags of the account.
            This allows for setting specific bits without knowledge of existing flags, you can also use
            :class:`stellar_sdk.operation.set_options.AuthorizationFlag`

            * AUTHORIZATION_REQUIRED = 1
            * AUTHORIZATION_REVOCABLE = 2
            * AUTHORIZATION_IMMUTABLE = 4
            * AUTHORIZATION_CLAWBACK_ENABLED = 8

        :param set_flags: Indicates which flags to set. For details about the flags,
            please refer to the `Control Access to an Asset - Flag <https://developers.stellar.org/docs/issuing-assets/control-asset-access/#flags>`__.
            The bit mask integer adds onto the existing flags of the account.
            This allows for setting specific bits without knowledge of existing flags, you can also use
            :class:`stellar_sdk.operation.set_options.AuthorizationFlag`

            * AUTHORIZATION_REQUIRED = 1
            * AUTHORIZATION_REVOCABLE = 2
            * AUTHORIZATION_IMMUTABLE = 4
            * AUTHORIZATION_CLAWBACK_ENABLED = 8

        :param master_weight: A number from 0-255 (inclusive) representing the weight of the master key.
            If the weight of the master key is updated to 0, it is effectively disabled.
        :param low_threshold: A number from 0-255 (inclusive) representing the threshold this account sets on all
            operations it performs that have `a low threshold <https://developers.stellar.org/docs/glossary/multisig/>`_.
        :param med_threshold: A number from 0-255 (inclusive) representing the threshold this account sets on all
            operations it performs that have `a medium threshold <https://developers.stellar.org/docs/glossary/multisig/>`_.
        :param high_threshold: A number from 0-255 (inclusive) representing the threshold this account sets on all
            operations it performs that have `a high threshold <https://developers.stellar.org/docs/glossary/multisig/>`_.
        :param home_domain: sets the home domain used for
            reverse `federation <https://developers.stellar.org/docs/glossary/federation/>`_ lookup.
        :param signer: Add, update, or remove a signer from the account.
        :param source: The source account for the operation. Defaults to the transaction's source account.
        :return: This builder instance.

        """
        op = SetOptions(
            inflation_dest=inflation_dest,
            clear_flags=clear_flags,
            set_flags=set_flags,
            master_weight=master_weight,
            low_threshold=low_threshold,
            med_threshold=med_threshold,
            high_threshold=high_threshold,
            signer=signer,
            home_domain=home_domain,
            source=source,
        )
        return self.append_operation(op)

    def append_ed25519_public_key_signer(
        self,
        account_id: str,
        weight: int,
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> "TransactionBuilder":
        """Add a ed25519 public key signer to an account via a :class:`SetOptions
        <stellar_sdk.operation.SetOptions` operation. This is a helper
        function for :meth:`append_set_options_op`.

        :param account_id: The account id of the new ed25519_public_key signer.
            (ex. ``"GDNA2V62PVEFBZ74CDJKTUHLY4Y7PL5UAV2MAM4VWF6USFE3SH2354AD"``)
        :param weight: The weight of the new signer.
        :param source: The source account that is adding a signer to its
            list of signers.
        :return: This builder instance.
        """
        signer = Signer.ed25519_public_key(account_id=account_id, weight=weight)
        return self.append_set_options_op(signer=signer, source=source)

    def append_hashx_signer(
        self,
        sha256_hash: Union[bytes, str],
        weight: int,
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> "TransactionBuilder":
        """Add a sha256 hash(HashX) signer to an account via a :class:`SetOptions
        <stellar_sdk.operation.SetOptions` operation. This is a helper
        function for :meth:`append_set_options_op`.

        :param sha256_hash: The address of the new sha256 hash(hashX) signer,
            a 32 byte hash, hex encoded string or encode strkey.
            (ex. ``"XDNA2V62PVEFBZ74CDJKTUHLY4Y7PL5UAV2MAM4VWF6USFE3SH235FXL"``,
            ``"da0d57da7d4850e7fc10d2a9d0ebc731f7afb40574c03395b17d49149b91f5be"`` or bytes)
        :param weight: The weight of the new signer.
        :param source: The source account that is adding a signer to its
            list of signers.
        :return: This builder instance.
        """
        if isinstance(sha256_hash, str) and is_valid_hash(sha256_hash):
            sha256_hash = hex_to_bytes(sha256_hash)
        signer = Signer.sha256_hash(sha256_hash=sha256_hash, weight=weight)
        return self.append_set_options_op(signer=signer, source=source)

    def append_pre_auth_tx_signer(
        self, pre_auth_tx_hash: Union[str, bytes], weight: int, source=None
    ) -> "TransactionBuilder":
        """Add a PreAuthTx signer to an account via a :class:`SetOptions <stellar_sdk.operation.SetOptions` operation.
        This is a helper function for :meth:`append_set_options_op`.

        :param pre_auth_tx_hash: The address of the new preAuthTx signer - obtained by calling `hash` on
            the :class:`TransactionEnvelope <stellar_sdk.transaction_envelope.TransactionEnvelope>`,
            a 32 byte hash, hex encoded string or encode strkey.
            (ex. ``"TDNA2V62PVEFBZ74CDJKTUHLY4Y7PL5UAV2MAM4VWF6USFE3SH234BSS"``,
            ``"da0d57da7d4850e7fc10d2a9d0ebc731f7afb40574c03395b17d49149b91f5be"`` or bytes)
        :param weight: The weight of the new signer.
        :param source: The source account that is adding a signer to its
            list of signers.
        :return: This builder instance.
        """
        if isinstance(pre_auth_tx_hash, str) and is_valid_hash(pre_auth_tx_hash):
            pre_auth_tx_hash = hex_to_bytes(pre_auth_tx_hash)
        signer = Signer.pre_auth_tx(pre_auth_tx_hash=pre_auth_tx_hash, weight=weight)
        return self.append_set_options_op(signer=signer, source=source)

    def append_manage_buy_offer_op(
        self,
        selling: Asset,
        buying: Asset,
        amount: Union[str, Decimal],
        price: Union[Price, str, Decimal],
        offer_id: int = 0,
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> "TransactionBuilder":
        """Append a :class:`ManageBuyOffer <stellar_sdk.operation.ManageBuyOffer>`
        operation to the list of operations.

        :param selling: What you're selling.
        :param buying: What you're buying.
        :param amount: Amount being bought. if set to ``0``, delete the offer.
        :param price: Price of thing being bought in terms of what you are selling.
        :param offer_id: If ``0``, will create a new offer (default). Otherwise,
            edits an existing offer.
        :param source: The source account for the operation. Defaults to the transaction's source account.
        :return: This builder instance.
        """

        op = ManageBuyOffer(
            selling=selling,
            buying=buying,
            amount=amount,
            price=price,
            offer_id=offer_id,
            source=source,
        )
        return self.append_operation(op)

    def append_manage_sell_offer_op(
        self,
        selling: Asset,
        buying: Asset,
        amount: Union[str, Decimal],
        price: Union[Price, str, Decimal],
        offer_id: int = 0,
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> "TransactionBuilder":
        """Append a :class:`ManageSellOffer <stellar_sdk.operation.ManageSellOffer>`
        operation to the list of operations.

        :param selling: What you're selling.
        :param buying: What you're buying.
        :param amount: The total amount you're selling. If ``0``, deletes the offer.
        :param price: Price of 1 unit of `selling` in terms of `buying`.
        :param offer_id: If ``0``, will create a new offer (default). Otherwise,
            edits an existing offer.
        :param source: The source account for the operation. Defaults to the transaction's source account.
        :return: This builder instance.
        """

        op = ManageSellOffer(
            selling=selling,
            buying=buying,
            amount=amount,
            price=price,
            offer_id=offer_id,
            source=source,
        )
        return self.append_operation(op)

    def append_create_passive_sell_offer_op(
        self,
        selling: Asset,
        buying: Asset,
        amount: Union[str, Decimal],
        price: Union[Price, str, Decimal],
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> "TransactionBuilder":
        """Append a :class:`CreatePassiveSellOffer
        <stellar_sdk.operation.CreatePassiveSellOffer>` operation to the list of
        operations.

        :param selling: What you're selling.
        :param buying: What you're buying.
        :param amount: The total amount you're selling.
        :param price: Price of 1 unit of `selling` in terms of `buying`.
        :param source: The source account for the operation. Defaults to the transaction's source account.
        :return: This builder instance.
        """

        op = CreatePassiveSellOffer(selling, buying, amount, price, source)
        return self.append_operation(op)

    def append_account_merge_op(
        self,
        destination: Union[MuxedAccount, str],
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> "TransactionBuilder":
        """Append a :class:`AccountMerge
        <stellar_sdk.operation.AccountMerge>` operation to the list of
        operations.

        :param destination: The ID of the offer. 0 for new offer. Set to
            existing offer ID to update or delete.
        :param source: The source address that is being merged into the
            destination account.
        :return: This builder instance.

        """
        op = AccountMerge(destination, source)
        return self.append_operation(op)

    def append_inflation_op(
        self, source: Optional[Union[MuxedAccount, str]] = None
    ) -> "TransactionBuilder":
        """Append a :class:`Inflation
        <stellar_sdk.operation.Inflation>` operation to the list of
        operations.

        :param source: The source address that is running the inflation
            operation.
        :return: This builder instance.

        """
        op = Inflation(source)
        return self.append_operation(op)

    def append_manage_data_op(
        self,
        data_name: str,
        data_value: Union[str, bytes, None],
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> "TransactionBuilder":
        """Append a :class:`ManageData <stellar_sdk.operation.ManageData>`
        operation to the list of operations.

        :param data_name: If this is a new Name
            it will add the given name/value pair to the account. If this Name
            is already present then the associated value will be modified. Up to 64 bytes long.
        :param data_value: If not present then the existing `data_name` will be deleted.
            If present then this value will be set in the DataEntry. Up to 64 bytes long.
        :param source: The source account on which data is being managed.
            operation.
        :return: This builder instance.

        """
        op = ManageData(data_name, data_value, source)
        return self.append_operation(op)

    def append_bump_sequence_op(
        self, bump_to: int, source: Optional[Union[MuxedAccount, str]] = None
    ) -> "TransactionBuilder":
        """Append a :class:`BumpSequence <stellar_sdk.operation.BumpSequence>`
        operation to the list of operations.

        :param bump_to: Sequence number to bump to.
        :param source: The source address that is running the inflation operation.
        :return: This builder instance.

        """
        op = BumpSequence(bump_to, source)
        return self.append_operation(op)

    def append_create_claimable_balance_op(
        self,
        asset: Asset,
        amount: Union[str, Decimal],
        claimants: Sequence[Claimant],
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> "TransactionBuilder":
        """Append a :class:`CreateClaimableBalance <stellar_sdk.operation.CreateClaimableBalance>`
        operation to the list of operations.

        :param asset: The asset for the claimable balance.
        :param amount: the amount of the asset.
        :param claimants: A list of Claimants.
        :param source: The source account for the operation. Defaults to the transaction's source account.
        """
        op = CreateClaimableBalance(asset, amount, claimants, source)
        return self.append_operation(op)

    def append_claim_claimable_balance_op(
        self, balance_id: str, source: Optional[Union[MuxedAccount, str]] = None
    ) -> "TransactionBuilder":
        """Append a :class:`ClaimClaimableBalance <stellar_sdk.operation.ClaimClaimableBalance>`
        operation to the list of operations.

        :param balance_id: The claimable balance id to be claimed.
        :param source: The source account for the operation. Defaults to the transaction's source account.
        """
        op = ClaimClaimableBalance(balance_id, source)
        return self.append_operation(op)

    def append_begin_sponsoring_future_reserves_op(
        self, sponsored_id: str, source: Optional[Union[MuxedAccount, str]] = None
    ) -> "TransactionBuilder":
        """Append a :class:`BeginSponsoringFutureReserves <stellar_sdk.operation.BeginSponsoringFutureReserves>`
        operation to the list of operations.

        :param sponsored_id: The sponsored account id.
        :param source: The source account for the operation. Defaults to the transaction's source account.
        :return: This builder instance.
        """
        op = BeginSponsoringFutureReserves(sponsored_id, source)
        return self.append_operation(op)

    def append_end_sponsoring_future_reserves_op(
        self, source: Optional[Union[MuxedAccount, str]] = None
    ) -> "TransactionBuilder":
        """Append a :class:`EndSponsoringFutureReserves <stellar_sdk.operation.EndSponsoringFutureReserves>`
        operation to the list of operations.

        :param source: The source account for the operation. Defaults to the transaction's source account.
        :return: This builder instance.
        """
        op = EndSponsoringFutureReserves(source)
        return self.append_operation(op)

    def append_revoke_account_sponsorship_op(
        self, account_id: str, source: Optional[Union[MuxedAccount, str]] = None
    ) -> "TransactionBuilder":
        """Append a :class:`RevokeSponsorship <stellar_sdk.operation.RevokeSponsorship>` operation
        for an account to the list of operations.

        :param account_id: The sponsored account ID.
        :param source: The source account for the operation. Defaults to the transaction's source account.
        :return: This builder instance.
        """
        op = RevokeSponsorship.revoke_account_sponsorship(account_id, source)
        return self.append_operation(op)

    def append_revoke_trustline_sponsorship_op(
        self,
        account_id: str,
        asset: Union[Asset, LiquidityPoolId],
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> "TransactionBuilder":
        """Append a :class:`RevokeSponsorship <stellar_sdk.operation.RevokeSponsorship>` operation
        for a trustline to the list of operations.

        :param account_id: The account ID which owns the trustline.
        :param asset: The asset in the trustline.
        :param source: The source account for the operation. Defaults to the transaction's source account.
        :return: This builder instance.
        """
        op = RevokeSponsorship.revoke_trustline_sponsorship(account_id, asset, source)
        return self.append_operation(op)

    def append_revoke_offer_sponsorship_op(
        self,
        seller_id: str,
        offer_id: int,
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> "TransactionBuilder":
        """Append a :class:`RevokeSponsorship <stellar_sdk.operation.RevokeSponsorship>` operation
        for an offer to the list of operations.

        :param seller_id: The account ID which created the offer.
        :param offer_id: The offer ID.
        :param source: The source account for the operation. Defaults to the transaction's source account.
        :return: This builder instance.
        """
        op = RevokeSponsorship.revoke_offer_sponsorship(seller_id, offer_id, source)
        return self.append_operation(op)

    def append_revoke_data_sponsorship_op(
        self,
        account_id: str,
        data_name: str,
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> "TransactionBuilder":
        """Append a :class:`RevokeSponsorship <stellar_sdk.operation.RevokeSponsorship>` operation
        for a data entry to the list of operations.

        :param account_id: The account ID which owns the data entry.
        :param data_name: The name of the data entry
        :param source: The source account for the operation. Defaults to the transaction's source account.
        :return: This builder instance.
        """
        op = RevokeSponsorship.revoke_data_sponsorship(account_id, data_name, source)
        return self.append_operation(op)

    def append_revoke_claimable_balance_sponsorship_op(
        self,
        claimable_balance_id: str,
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> "TransactionBuilder":
        """Append a :class:`RevokeSponsorship <stellar_sdk.operation.RevokeSponsorship>` operation
        for a claimable to the list of operations.

        :param claimable_balance_id: The sponsored claimable balance ID.
        :param source: The source account for the operation. Defaults to the transaction's source account.
        :return: This builder instance.
        """
        op = RevokeSponsorship.revoke_claimable_balance_sponsorship(
            claimable_balance_id, source
        )
        return self.append_operation(op)

    def append_revoke_liquidity_pool_sponsorship_op(
        self,
        liquidity_pool_id: str,
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> "TransactionBuilder":
        """Append a :class:`RevokeSponsorship <stellar_sdk.operation.RevokeSponsorship>` operation
        for a claimable to the list of operations.

        :param liquidity_pool_id: The sponsored liquidity pool ID in hex string.
        :param source: The source account for the operation. Defaults to the transaction's source account.
        :return: This builder instance.
        """
        op = RevokeSponsorship.revoke_liquidity_pool_sponsorship(
            liquidity_pool_id, source
        )
        return self.append_operation(op)

    def append_revoke_ed25519_public_key_signer_sponsorship_op(
        self,
        account_id: str,
        signer_key: str,
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> "TransactionBuilder":
        """Append a :class:`RevokeSponsorship <stellar_sdk.operation.RevokeSponsorship>` operation
        for an ed25519_public_key signer to the list of operations.

        :param account_id: The account ID where the signer sponsorship is being removed from.
        :param signer_key: The account id of the ed25519_public_key signer.
            (ex. ``"GDNA2V62PVEFBZ74CDJKTUHLY4Y7PL5UAV2MAM4VWF6USFE3SH2354AD"``)
        :param source: The source account for the operation. Defaults to the transaction's source account.
        :return: This builder instance.
        """
        key = SignerKey.ed25519_public_key(signer_key)
        op = RevokeSponsorship.revoke_signer_sponsorship(account_id, key, source)
        return self.append_operation(op)

    def append_revoke_hashx_signer_sponsorship_op(
        self,
        account_id: str,
        signer_key: Union[bytes, str],
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> "TransactionBuilder":
        """Append a :class:`RevokeSponsorship <stellar_sdk.operation.RevokeSponsorship>` operation
        for a hashx signer to the list of operations.

        :param account_id: The account ID where the signer sponsorship is being removed from.
        :param signer_key: The account id of the hashx signer.
            (ex. ``"XDNA2V62PVEFBZ74CDJKTUHLY4Y7PL5UAV2MAM4VWF6USFE3SH235FXL"``,
            ``"da0d57da7d4850e7fc10d2a9d0ebc731f7afb40574c03395b17d49149b91f5be"`` or bytes)
        :param source: The source account for the operation. Defaults to the transaction's source account.
        :return: This builder instance.
        """
        if isinstance(signer_key, str) and is_valid_hash(signer_key):
            signer_key = hex_to_bytes(signer_key)
        key = SignerKey.sha256_hash(signer_key)
        op = RevokeSponsorship.revoke_signer_sponsorship(account_id, key, source)
        return self.append_operation(op)

    def append_revoke_pre_auth_tx_signer_sponsorship_op(
        self,
        account_id: str,
        signer_key: Union[bytes, str],
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> "TransactionBuilder":
        """Append a :class:`RevokeSponsorship <stellar_sdk.operation.RevokeSponsorship>` operation
        for a pre_auth_tx signer to the list of operations.

        :param account_id: The account ID where the signer sponsorship is being removed from.
        :param signer_key: The account id of the pre_auth_tx signer.
            (ex. ``"TDNA2V62PVEFBZ74CDJKTUHLY4Y7PL5UAV2MAM4VWF6USFE3SH234BSS"``,
            ``"da0d57da7d4850e7fc10d2a9d0ebc731f7afb40574c03395b17d49149b91f5be"`` or bytes)
        :param source: The source account for the operation. Defaults to the transaction's source account.
        :return: This builder instance.
        """
        if isinstance(signer_key, str) and is_valid_hash(signer_key):
            signer_key = hex_to_bytes(signer_key)
        key = SignerKey.pre_auth_tx(signer_key)
        op = RevokeSponsorship.revoke_signer_sponsorship(account_id, key, source)
        return self.append_operation(op)

    def append_clawback_op(
        self,
        asset: Asset,
        from_: Union[MuxedAccount, str],
        amount: Union[str, Decimal],
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> "TransactionBuilder":
        """Append an :class:`Clawback <stellar_sdk.operation.Clawback>`
        operation to the list of operations.

        :param asset: The asset being clawed back.
        :param from_: The public key of the account to claw back from.
        :param amount: The amount of the asset to claw back.
        :param source: The source account for the operation. Defaults to the
            transaction's source account.
        """
        op = Clawback(asset, from_, amount, source)
        return self.append_operation(op)

    def append_clawback_claimable_balance_op(
        self, balance_id: str, source: Optional[Union[MuxedAccount, str]] = None
    ) -> "TransactionBuilder":
        """Append an :class:`ClawbackClaimableBalance <stellar_sdk.operation.ClawbackClaimableBalance>`
        operation to the list of operations.

        :param balance_id: The claimable balance ID to be clawed back.
        :param source: The source account for the operation. Defaults to the
            transaction's source account.
        :return: This builder instance.
        """
        op = ClawbackClaimableBalance(balance_id, source)
        return self.append_operation(op)

    def append_set_trust_line_flags_op(
        self,
        trustor: str,
        asset: Asset,
        clear_flags: TrustLineFlags = None,
        set_flags: TrustLineFlags = None,
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> "TransactionBuilder":
        """Append an :class:`SetTrustLineFlags <stellar_sdk.operation.SetTrustLineFlags>`
        operation to the list of operations.

        :param trustor: The account whose trustline this is.
        :param asset: The asset on the trustline.
        :param clear_flags: The flags to clear.
        :param set_flags: The flags to set.
        :param source: The source account for the operation. Defaults to the
            transaction's source account.
        :return: This builder instance.
        """
        op = SetTrustLineFlags(trustor, asset, clear_flags, set_flags, source)
        return self.append_operation(op)

    def append_liquidity_pool_deposit_op(
        self,
        liquidity_pool_id: str,
        max_amount_a: Union[str, Decimal],
        max_amount_b: Union[str, Decimal],
        min_price: Union[str, Decimal, Price],
        max_price: Union[str, Decimal, Price],
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> "TransactionBuilder":
        """Append an :class:`LiquidityPoolDeposit <stellar_sdk.operation.LiquidityPoolDeposit>`
        operation to the list of operations.

        :param liquidity_pool_id: The liquidity pool ID.
        :param max_amount_a: Maximum amount of first asset to deposit.
        :param max_amount_b: Maximum amount of second asset to deposit.
        :param min_price: Minimum deposit_a/deposit_b price.
        :param max_price: Maximum deposit_a/deposit_b price.
        :param source: The source account for the operation. Defaults to the
            transaction's source account.
        :return: This builder instance.
        """
        op = LiquidityPoolDeposit(
            liquidity_pool_id, max_amount_a, max_amount_b, min_price, max_price, source
        )
        return self.append_operation(op)

    def append_liquidity_pool_withdraw_op(
        self,
        liquidity_pool_id: str,
        amount: Union[str, Decimal],
        min_amount_a: Union[str, Decimal],
        min_amount_b: Union[str, Decimal],
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> "TransactionBuilder":
        """Append an :class:`LiquidityPoolWithdraw <stellar_sdk.operation.LiquidityPoolWithdraw>`
        operation to the list of operations.

        :param liquidity_pool_id: The liquidity pool ID.
        :param amount: Amount of pool shares to withdraw.
        :param min_amount_a: Minimum amount of first asset to withdraw.
        :param min_amount_b: Minimum amount of second asset to withdraw.
        :param source: The source account for the operation. Defaults to the
            transaction's source account.
        :return: This builder instance.
        """
        op = LiquidityPoolWithdraw(
            liquidity_pool_id, amount, min_amount_a, min_amount_b, source
        )
        return self.append_operation(op)

    def append_invoke_contract_function_op(
        self,
        contract_id: str,
        function_name: str,
        parameters: Sequence[stellar_xdr.SCVal] = None,
        auth: Sequence[stellar_xdr.SorobanAuthorizationEntry] = None,
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> "TransactionBuilder":
        """Append an :class:`InvokeHostFunction <stellar_sdk.operation.InvokeHostFunction>` operation to the list of operations.

        You can use this method to invoke a contract function.

        :param contract_id: The ID of the contract to invoke.
        :param function_name: The name of the function to invoke.
        :param parameters: The parameters to pass to the method.
        :param auth: The authorizations required to execute the host function.
        :param source: The source account for the operation. Defaults to the
            transaction's source account.
        :return: This builder instance.
        """
        if not StrKey.is_valid_contract(contract_id):
            raise ValueError("`contract_id` is invalid.")

        if parameters is None:
            parameters = []

        host_function = stellar_xdr.HostFunction(
            stellar_xdr.HostFunctionType.HOST_FUNCTION_TYPE_INVOKE_CONTRACT,
            invoke_contract=stellar_xdr.InvokeContractArgs(
                contract_address=Address(contract_id).to_xdr_sc_address(),
                function_name=stellar_xdr.SCSymbol(
                    sc_symbol=function_name.encode("utf-8")
                ),
                args=list(parameters),
            ),
        )
        op = InvokeHostFunction(host_function=host_function, auth=auth, source=source)
        return self.append_operation(op)

    def append_upload_contract_wasm_op(
        self,
        contract: Union[bytes, str],
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> "TransactionBuilder":
        """Append an :class:`InvokeHostFunction <stellar_sdk.operation.InvokeHostFunction>` operation to the list of operations.

        You can use this method to install a contract code,
        and then use :func:`append_create_contract_op` to create a contract.

        :param contract: The contract code to install, path to a file or bytes.
        :param source: The source account for the operation. Defaults to the
            transaction's source account.
        :return: This builder instance.
        """

        if isinstance(contract, str):
            with open(contract, "rb") as f:
                contract = f.read()

        host_function = stellar_xdr.HostFunction(
            stellar_xdr.HostFunctionType.HOST_FUNCTION_TYPE_UPLOAD_CONTRACT_WASM,
            wasm=contract,
        )
        op = InvokeHostFunction(host_function=host_function, auth=[], source=source)
        return self.append_operation(op)

    def append_create_contract_op(
        self,
        wasm_id: Union[bytes, str],
        address: Union[str, Address],
        constructor_args: Optional[Sequence[stellar_xdr.SCVal]] = None,
        salt: Optional[bytes] = None,
        auth: Sequence[stellar_xdr.SorobanAuthorizationEntry] = None,
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> "TransactionBuilder":
        """Append an :class:`InvokeHostFunction <stellar_sdk.operation.InvokeHostFunction>` operation to the list of operations.

        You can use this method to create a contract.

        :param wasm_id: The ID of the contract code to install.
        :param address: The address using to derive the contract ID.
        :param constructor_args: The optional parameters to pass to the constructor of this contract.
        :param salt: The 32-byte salt to use to derive the contract ID.
        :param auth: The authorizations required to execute the host function.
        :param source: The source account for the operation. Defaults to the
            transaction's source account.
        :return: This builder instance.
        """
        if isinstance(wasm_id, str):
            wasm_id = binascii.unhexlify(wasm_id)

        if salt is None:
            salt = os.urandom(32)
        else:
            if len(salt) != 32:
                raise ValueError("`salt` must be 32 bytes long")

        if isinstance(address, str):
            address = Address(address)

        create_contract = stellar_xdr.CreateContractArgsV2(
            contract_id_preimage=stellar_xdr.ContractIDPreimage(
                stellar_xdr.ContractIDPreimageType.CONTRACT_ID_PREIMAGE_FROM_ADDRESS,
                from_address=stellar_xdr.ContractIDPreimageFromAddress(
                    address=address.to_xdr_sc_address(),
                    salt=stellar_xdr.Uint256(salt),
                ),
            ),
            executable=stellar_xdr.ContractExecutable(
                stellar_xdr.ContractExecutableType.CONTRACT_EXECUTABLE_WASM,
                stellar_xdr.Hash(wasm_id),
            ),
            constructor_args=(
                list(constructor_args) if constructor_args is not None else []
            ),
        )

        host_function = stellar_xdr.HostFunction(
            stellar_xdr.HostFunctionType.HOST_FUNCTION_TYPE_CREATE_CONTRACT_V2,
            create_contract_v2=create_contract,
        )

        op = InvokeHostFunction(host_function=host_function, auth=auth, source=source)
        return self.append_operation(op)

    def append_create_stellar_asset_contract_from_asset_op(
        self,
        asset: Asset,
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> "TransactionBuilder":
        """Append an :class:`InvokeHostFunction <stellar_sdk.operation.InvokeHostFunction>` operation to the list of operations.

        You can use this method to deploy a contract that wraps a classic asset.

        :param asset: The asset to wrap.
        :param source: The source account for the operation. Defaults to the
            transaction's source account.
        :return: This builder instance.
        """
        asset_param = asset.to_xdr_object()

        create_contract = stellar_xdr.CreateContractArgs(
            contract_id_preimage=stellar_xdr.ContractIDPreimage(
                stellar_xdr.ContractIDPreimageType.CONTRACT_ID_PREIMAGE_FROM_ASSET,
                from_asset=asset_param,
            ),
            executable=stellar_xdr.ContractExecutable(
                stellar_xdr.ContractExecutableType.CONTRACT_EXECUTABLE_STELLAR_ASSET,
            ),
        )

        host_function = stellar_xdr.HostFunction(
            stellar_xdr.HostFunctionType.HOST_FUNCTION_TYPE_CREATE_CONTRACT,
            create_contract=create_contract,
        )

        op = InvokeHostFunction(host_function=host_function, auth=[], source=source)
        return self.append_operation(op)

    def append_create_stellar_asset_contract_from_address_op(
        self,
        address: Union[str, Address],
        salt: Optional[bytes] = None,
        auth: Sequence[stellar_xdr.SorobanAuthorizationEntry] = None,
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> "TransactionBuilder":
        """Append an :class:`InvokeHostFunction <stellar_sdk.operation.InvokeHostFunction>` operation to the list of operations.

        You can use this method to create a new Soroban token contract.

        I do not recommend using this method, please check
        `the documentation <https://soroban.stellar.org/docs/learn/faq#should-i-issue-my-token-as-a-stellar-asset-or-a-custom-soroban-token>`__ for more information.

        :param address: The address using to derive the contract ID.
        :param salt: The 32-byte salt to use to derive the contract ID.
        :param auth: The authorizations required to execute the host function.
        :param source: The source account for the operation. Defaults to the
            transaction's source account.
        :return: This builder instance.
        """
        if salt is None:
            salt = os.urandom(32)
        else:
            if len(salt) != 32:
                raise ValueError("`salt` must be 32 bytes long")

        if isinstance(address, str):
            address = Address(address)

        create_contract = stellar_xdr.CreateContractArgs(
            contract_id_preimage=stellar_xdr.ContractIDPreimage(
                stellar_xdr.ContractIDPreimageType.CONTRACT_ID_PREIMAGE_FROM_ADDRESS,
                from_address=stellar_xdr.ContractIDPreimageFromAddress(
                    address=address.to_xdr_sc_address(),
                    salt=stellar_xdr.Uint256(salt),
                ),
            ),
            executable=stellar_xdr.ContractExecutable(
                stellar_xdr.ContractExecutableType.CONTRACT_EXECUTABLE_STELLAR_ASSET,
            ),
        )

        host_function = stellar_xdr.HostFunction(
            stellar_xdr.HostFunctionType.HOST_FUNCTION_TYPE_CREATE_CONTRACT,
            create_contract=create_contract,
        )

        op = InvokeHostFunction(host_function=host_function, auth=auth, source=source)
        return self.append_operation(op)

    def append_extend_footprint_ttl_op(
        self, extend_to: int, source: Optional[Union[MuxedAccount, str]] = None
    ) -> "TransactionBuilder":
        """Append an :class:`ExtendFootprintTTL <stellar_sdk.operation.ExtendFootprintTTL>` operation to the list of operations.

        :param extend_to: The number of ledgers past the LCL (last closed ledger)
            by which to extend the validity of the ledger keys in this transaction.
        :param source: The source account for the operation. Defaults to the
            transaction's source account.
        :return: This builder instance.
        """
        op = ExtendFootprintTTL(extend_to=extend_to, source=source)
        return self.append_operation(op)

    def append_restore_footprint_op(
        self, source: Optional[Union[MuxedAccount, str]] = None
    ):
        """Append an :class:`RestoreFootprint <stellar_sdk.operation.RestoreFootprint>` operation to the list of operations.

        :param source: The source account for the operation. Defaults to the
            transaction's source account.
        :return: This builder instance.
        """
        op = RestoreFootprint(source)
        return self.append_operation(op)

    def __repr__(self):
        return (
            f"<TransactionBuilder [source_account={self.source_account}, "
            f"base_fee={self.base_fee}, network_passphrase={self.network_passphrase}, "
            f"operations={self.operations}, memo={self.memo}, "
            f"time_bounds={self.time_bounds}, "
            f"ledger_bounds={self.ledger_bounds}, "
            f"min_sequence_number={self.min_sequence_number}, "
            f"min_sequence_age={self.min_sequence_age}, "
            f"min_sequence_ledger_gap={self.min_sequence_ledger_gap}, "
            f"extra_signers={self.extra_signers}, "
            f"soroban_data={self.soroban_data}, "
            f"v1={self.v1}]>"
        )



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .sc_spec_type_def import SCSpecTypeDef

__all__ = ["SCSpecTypeMap"]


class SCSpecTypeMap:
    """
    XDR Source Code::

        struct SCSpecTypeMap
        {
            SCSpecTypeDef keyType;
            SCSpecTypeDef valueType;
        };
    """

    def __init__(
        self,
        key_type: SCSpecTypeDef,
        value_type: SCSpecTypeDef,
    ) -> None:
        self.key_type = key_type
        self.value_type = value_type

    def pack(self, packer: Packer) -> None:
        self.key_type.pack(packer)
        self.value_type.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SCSpecTypeMap:
        key_type = SCSpecTypeDef.unpack(unpacker)
        value_type = SCSpecTypeDef.unpack(unpacker)
        return cls(
            key_type=key_type,
            value_type=value_type,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SCSpecTypeMap:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SCSpecTypeMap:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.key_type,
                self.value_type,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.key_type == other.key_type and self.value_type == other.value_type

    def __repr__(self):
        out = [
            f"key_type={self.key_type}",
            f"value_type={self.value_type}",
        ]
        return f"<SCSpecTypeMap [{', '.join(out)}]>"



from .accounts_call_builder import *
from .assets_call_builder import *
from .base_call_builder import *
from .claimable_balances_call_builder import *
from .data_call_builder import *
from .effects_call_builder import *
from .fee_stats_call_builder import *
from .ledgers_call_builder import *
from .liquidity_pools_builder import *
from .offers_call_builder import *
from .operations_call_builder import *
from .orderbook_call_builder import *
from .payments_call_builder import *
from .root_call_builder import *
from .strict_receive_paths_call_builder import *
from .strict_send_paths_call_builder import *
from .trades_aggregation_call_builder import *
from .trades_call_builder import *
from .transactions_call_builder import *



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .base import Boolean
from .uint32 import Uint32

__all__ = ["TimeSlicedNodeData"]


class TimeSlicedNodeData:
    """
    XDR Source Code::

        struct TimeSlicedNodeData
        {
            uint32 addedAuthenticatedPeers;
            uint32 droppedAuthenticatedPeers;
            uint32 totalInboundPeerCount;
            uint32 totalOutboundPeerCount;

            // SCP stats
            uint32 p75SCPFirstToSelfLatencyMs;
            uint32 p75SCPSelfToOtherLatencyMs;

            // How many times the node lost sync in the time slice
            uint32 lostSyncCount;

            // Config data
            bool isValidator;
            uint32 maxInboundPeerCount;
            uint32 maxOutboundPeerCount;
        };
    """

    def __init__(
        self,
        added_authenticated_peers: Uint32,
        dropped_authenticated_peers: Uint32,
        total_inbound_peer_count: Uint32,
        total_outbound_peer_count: Uint32,
        p75_scp_first_to_self_latency_ms: Uint32,
        p75_scp_self_to_other_latency_ms: Uint32,
        lost_sync_count: Uint32,
        is_validator: bool,
        max_inbound_peer_count: Uint32,
        max_outbound_peer_count: Uint32,
    ) -> None:
        self.added_authenticated_peers = added_authenticated_peers
        self.dropped_authenticated_peers = dropped_authenticated_peers
        self.total_inbound_peer_count = total_inbound_peer_count
        self.total_outbound_peer_count = total_outbound_peer_count
        self.p75_scp_first_to_self_latency_ms = p75_scp_first_to_self_latency_ms
        self.p75_scp_self_to_other_latency_ms = p75_scp_self_to_other_latency_ms
        self.lost_sync_count = lost_sync_count
        self.is_validator = is_validator
        self.max_inbound_peer_count = max_inbound_peer_count
        self.max_outbound_peer_count = max_outbound_peer_count

    def pack(self, packer: Packer) -> None:
        self.added_authenticated_peers.pack(packer)
        self.dropped_authenticated_peers.pack(packer)
        self.total_inbound_peer_count.pack(packer)
        self.total_outbound_peer_count.pack(packer)
        self.p75_scp_first_to_self_latency_ms.pack(packer)
        self.p75_scp_self_to_other_latency_ms.pack(packer)
        self.lost_sync_count.pack(packer)
        Boolean(self.is_validator).pack(packer)
        self.max_inbound_peer_count.pack(packer)
        self.max_outbound_peer_count.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> TimeSlicedNodeData:
        added_authenticated_peers = Uint32.unpack(unpacker)
        dropped_authenticated_peers = Uint32.unpack(unpacker)
        total_inbound_peer_count = Uint32.unpack(unpacker)
        total_outbound_peer_count = Uint32.unpack(unpacker)
        p75_scp_first_to_self_latency_ms = Uint32.unpack(unpacker)
        p75_scp_self_to_other_latency_ms = Uint32.unpack(unpacker)
        lost_sync_count = Uint32.unpack(unpacker)
        is_validator = Boolean.unpack(unpacker)
        max_inbound_peer_count = Uint32.unpack(unpacker)
        max_outbound_peer_count = Uint32.unpack(unpacker)
        return cls(
            added_authenticated_peers=added_authenticated_peers,
            dropped_authenticated_peers=dropped_authenticated_peers,
            total_inbound_peer_count=total_inbound_peer_count,
            total_outbound_peer_count=total_outbound_peer_count,
            p75_scp_first_to_self_latency_ms=p75_scp_first_to_self_latency_ms,
            p75_scp_self_to_other_latency_ms=p75_scp_self_to_other_latency_ms,
            lost_sync_count=lost_sync_count,
            is_validator=is_validator,
            max_inbound_peer_count=max_inbound_peer_count,
            max_outbound_peer_count=max_outbound_peer_count,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> TimeSlicedNodeData:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> TimeSlicedNodeData:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.added_authenticated_peers,
                self.dropped_authenticated_peers,
                self.total_inbound_peer_count,
                self.total_outbound_peer_count,
                self.p75_scp_first_to_self_latency_ms,
                self.p75_scp_self_to_other_latency_ms,
                self.lost_sync_count,
                self.is_validator,
                self.max_inbound_peer_count,
                self.max_outbound_peer_count,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.added_authenticated_peers == other.added_authenticated_peers
            and self.dropped_authenticated_peers == other.dropped_authenticated_peers
            and self.total_inbound_peer_count == other.total_inbound_peer_count
            and self.total_outbound_peer_count == other.total_outbound_peer_count
            and self.p75_scp_first_to_self_latency_ms
            == other.p75_scp_first_to_self_latency_ms
            and self.p75_scp_self_to_other_latency_ms
            == other.p75_scp_self_to_other_latency_ms
            and self.lost_sync_count == other.lost_sync_count
            and self.is_validator == other.is_validator
            and self.max_inbound_peer_count == other.max_inbound_peer_count
            and self.max_outbound_peer_count == other.max_outbound_peer_count
        )

    def __repr__(self):
        out = [
            f"added_authenticated_peers={self.added_authenticated_peers}",
            f"dropped_authenticated_peers={self.dropped_authenticated_peers}",
            f"total_inbound_peer_count={self.total_inbound_peer_count}",
            f"total_outbound_peer_count={self.total_outbound_peer_count}",
            f"p75_scp_first_to_self_latency_ms={self.p75_scp_first_to_self_latency_ms}",
            f"p75_scp_self_to_other_latency_ms={self.p75_scp_self_to_other_latency_ms}",
            f"lost_sync_count={self.lost_sync_count}",
            f"is_validator={self.is_validator}",
            f"max_inbound_peer_count={self.max_inbound_peer_count}",
            f"max_outbound_peer_count={self.max_outbound_peer_count}",
        ]
        return f"<TimeSlicedNodeData [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .time_sliced_node_data import TimeSlicedNodeData
from .time_sliced_peer_data_list import TimeSlicedPeerDataList

__all__ = ["TopologyResponseBodyV2"]


class TopologyResponseBodyV2:
    """
    XDR Source Code::

        struct TopologyResponseBodyV2
        {
            TimeSlicedPeerDataList inboundPeers;
            TimeSlicedPeerDataList outboundPeers;
            TimeSlicedNodeData nodeData;
        };
    """

    def __init__(
        self,
        inbound_peers: TimeSlicedPeerDataList,
        outbound_peers: TimeSlicedPeerDataList,
        node_data: TimeSlicedNodeData,
    ) -> None:
        self.inbound_peers = inbound_peers
        self.outbound_peers = outbound_peers
        self.node_data = node_data

    def pack(self, packer: Packer) -> None:
        self.inbound_peers.pack(packer)
        self.outbound_peers.pack(packer)
        self.node_data.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> TopologyResponseBodyV2:
        inbound_peers = TimeSlicedPeerDataList.unpack(unpacker)
        outbound_peers = TimeSlicedPeerDataList.unpack(unpacker)
        node_data = TimeSlicedNodeData.unpack(unpacker)
        return cls(
            inbound_peers=inbound_peers,
            outbound_peers=outbound_peers,
            node_data=node_data,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> TopologyResponseBodyV2:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> TopologyResponseBodyV2:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.inbound_peers,
                self.outbound_peers,
                self.node_data,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.inbound_peers == other.inbound_peers
            and self.outbound_peers == other.outbound_peers
            and self.node_data == other.node_data
        )

    def __repr__(self):
        out = [
            f"inbound_peers={self.inbound_peers}",
            f"outbound_peers={self.outbound_peers}",
            f"node_data={self.node_data}",
        ]
        return f"<TopologyResponseBodyV2 [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .curve25519_public import Curve25519Public
from .node_id import NodeID
from .survey_message_command_type import SurveyMessageCommandType
from .uint32 import Uint32

__all__ = ["SurveyRequestMessage"]


class SurveyRequestMessage:
    """
    XDR Source Code::

        struct SurveyRequestMessage
        {
            NodeID surveyorPeerID;
            NodeID surveyedPeerID;
            uint32 ledgerNum;
            Curve25519Public encryptionKey;
            SurveyMessageCommandType commandType;
        };
    """

    def __init__(
        self,
        surveyor_peer_id: NodeID,
        surveyed_peer_id: NodeID,
        ledger_num: Uint32,
        encryption_key: Curve25519Public,
        command_type: SurveyMessageCommandType,
    ) -> None:
        self.surveyor_peer_id = surveyor_peer_id
        self.surveyed_peer_id = surveyed_peer_id
        self.ledger_num = ledger_num
        self.encryption_key = encryption_key
        self.command_type = command_type

    def pack(self, packer: Packer) -> None:
        self.surveyor_peer_id.pack(packer)
        self.surveyed_peer_id.pack(packer)
        self.ledger_num.pack(packer)
        self.encryption_key.pack(packer)
        self.command_type.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SurveyRequestMessage:
        surveyor_peer_id = NodeID.unpack(unpacker)
        surveyed_peer_id = NodeID.unpack(unpacker)
        ledger_num = Uint32.unpack(unpacker)
        encryption_key = Curve25519Public.unpack(unpacker)
        command_type = SurveyMessageCommandType.unpack(unpacker)
        return cls(
            surveyor_peer_id=surveyor_peer_id,
            surveyed_peer_id=surveyed_peer_id,
            ledger_num=ledger_num,
            encryption_key=encryption_key,
            command_type=command_type,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SurveyRequestMessage:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SurveyRequestMessage:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.surveyor_peer_id,
                self.surveyed_peer_id,
                self.ledger_num,
                self.encryption_key,
                self.command_type,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.surveyor_peer_id == other.surveyor_peer_id
            and self.surveyed_peer_id == other.surveyed_peer_id
            and self.ledger_num == other.ledger_num
            and self.encryption_key == other.encryption_key
            and self.command_type == other.command_type
        )

    def __repr__(self):
        out = [
            f"surveyor_peer_id={self.surveyor_peer_id}",
            f"surveyed_peer_id={self.surveyed_peer_id}",
            f"ledger_num={self.ledger_num}",
            f"encryption_key={self.encryption_key}",
            f"command_type={self.command_type}",
        ]
        return f"<SurveyRequestMessage [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .base import Opaque
from .binary_fuse_filter_type import BinaryFuseFilterType
from .short_hash_seed import ShortHashSeed
from .uint32 import Uint32

__all__ = ["SerializedBinaryFuseFilter"]


class SerializedBinaryFuseFilter:
    """
    XDR Source Code::

        struct SerializedBinaryFuseFilter
        {
            BinaryFuseFilterType type;

            // Seed used to hash input to filter
            ShortHashSeed inputHashSeed;

            // Seed used for internal filter hash operations
            ShortHashSeed filterSeed;
            uint32 segmentLength;
            uint32 segementLengthMask;
            uint32 segmentCount;
            uint32 segmentCountLength;
            uint32 fingerprintLength; // Length in terms of element count, not bytes

            // Array of uint8_t, uint16_t, or uint32_t depending on filter type
            opaque fingerprints<>;
        };
    """

    def __init__(
        self,
        type: BinaryFuseFilterType,
        input_hash_seed: ShortHashSeed,
        filter_seed: ShortHashSeed,
        segment_length: Uint32,
        segement_length_mask: Uint32,
        segment_count: Uint32,
        segment_count_length: Uint32,
        fingerprint_length: Uint32,
        fingerprints: bytes,
    ) -> None:
        self.type = type
        self.input_hash_seed = input_hash_seed
        self.filter_seed = filter_seed
        self.segment_length = segment_length
        self.segement_length_mask = segement_length_mask
        self.segment_count = segment_count
        self.segment_count_length = segment_count_length
        self.fingerprint_length = fingerprint_length
        self.fingerprints = fingerprints

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        self.input_hash_seed.pack(packer)
        self.filter_seed.pack(packer)
        self.segment_length.pack(packer)
        self.segement_length_mask.pack(packer)
        self.segment_count.pack(packer)
        self.segment_count_length.pack(packer)
        self.fingerprint_length.pack(packer)
        Opaque(self.fingerprints, 4294967295, False).pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SerializedBinaryFuseFilter:
        type = BinaryFuseFilterType.unpack(unpacker)
        input_hash_seed = ShortHashSeed.unpack(unpacker)
        filter_seed = ShortHashSeed.unpack(unpacker)
        segment_length = Uint32.unpack(unpacker)
        segement_length_mask = Uint32.unpack(unpacker)
        segment_count = Uint32.unpack(unpacker)
        segment_count_length = Uint32.unpack(unpacker)
        fingerprint_length = Uint32.unpack(unpacker)
        fingerprints = Opaque.unpack(unpacker, 4294967295, False)
        return cls(
            type=type,
            input_hash_seed=input_hash_seed,
            filter_seed=filter_seed,
            segment_length=segment_length,
            segement_length_mask=segement_length_mask,
            segment_count=segment_count,
            segment_count_length=segment_count_length,
            fingerprint_length=fingerprint_length,
            fingerprints=fingerprints,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SerializedBinaryFuseFilter:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SerializedBinaryFuseFilter:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.input_hash_seed,
                self.filter_seed,
                self.segment_length,
                self.segement_length_mask,
                self.segment_count,
                self.segment_count_length,
                self.fingerprint_length,
                self.fingerprints,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.type == other.type
            and self.input_hash_seed == other.input_hash_seed
            and self.filter_seed == other.filter_seed
            and self.segment_length == other.segment_length
            and self.segement_length_mask == other.segement_length_mask
            and self.segment_count == other.segment_count
            and self.segment_count_length == other.segment_count_length
            and self.fingerprint_length == other.fingerprint_length
            and self.fingerprints == other.fingerprints
        )

    def __repr__(self):
        out = [
            f"type={self.type}",
            f"input_hash_seed={self.input_hash_seed}",
            f"filter_seed={self.filter_seed}",
            f"segment_length={self.segment_length}",
            f"segement_length_mask={self.segement_length_mask}",
            f"segment_count={self.segment_count}",
            f"segment_count_length={self.segment_count_length}",
            f"fingerprint_length={self.fingerprint_length}",
            f"fingerprints={self.fingerprints}",
        ]
        return f"<SerializedBinaryFuseFilter [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .int64 import Int64
from .uint32 import Uint32

__all__ = ["StateArchivalSettings"]


class StateArchivalSettings:
    """
    XDR Source Code::

        struct StateArchivalSettings {
            uint32 maxEntryTTL;
            uint32 minTemporaryTTL;
            uint32 minPersistentTTL;

            // rent_fee = wfee_rate_average / rent_rate_denominator_for_type
            int64 persistentRentRateDenominator;
            int64 tempRentRateDenominator;

            // max number of entries that emit archival meta in a single ledger
            uint32 maxEntriesToArchive;

            // Number of snapshots to use when calculating average BucketList size
            uint32 bucketListSizeWindowSampleSize;

            // How often to sample the BucketList size for the average, in ledgers
            uint32 bucketListWindowSamplePeriod;

            // Maximum number of bytes that we scan for eviction per ledger
            uint32 evictionScanSize;

            // Lowest BucketList level to be scanned to evict entries
            uint32 startingEvictionScanLevel;
        };
    """

    def __init__(
        self,
        max_entry_ttl: Uint32,
        min_temporary_ttl: Uint32,
        min_persistent_ttl: Uint32,
        persistent_rent_rate_denominator: Int64,
        temp_rent_rate_denominator: Int64,
        max_entries_to_archive: Uint32,
        bucket_list_size_window_sample_size: Uint32,
        bucket_list_window_sample_period: Uint32,
        eviction_scan_size: Uint32,
        starting_eviction_scan_level: Uint32,
    ) -> None:
        self.max_entry_ttl = max_entry_ttl
        self.min_temporary_ttl = min_temporary_ttl
        self.min_persistent_ttl = min_persistent_ttl
        self.persistent_rent_rate_denominator = persistent_rent_rate_denominator
        self.temp_rent_rate_denominator = temp_rent_rate_denominator
        self.max_entries_to_archive = max_entries_to_archive
        self.bucket_list_size_window_sample_size = bucket_list_size_window_sample_size
        self.bucket_list_window_sample_period = bucket_list_window_sample_period
        self.eviction_scan_size = eviction_scan_size
        self.starting_eviction_scan_level = starting_eviction_scan_level

    def pack(self, packer: Packer) -> None:
        self.max_entry_ttl.pack(packer)
        self.min_temporary_ttl.pack(packer)
        self.min_persistent_ttl.pack(packer)
        self.persistent_rent_rate_denominator.pack(packer)
        self.temp_rent_rate_denominator.pack(packer)
        self.max_entries_to_archive.pack(packer)
        self.bucket_list_size_window_sample_size.pack(packer)
        self.bucket_list_window_sample_period.pack(packer)
        self.eviction_scan_size.pack(packer)
        self.starting_eviction_scan_level.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> StateArchivalSettings:
        max_entry_ttl = Uint32.unpack(unpacker)
        min_temporary_ttl = Uint32.unpack(unpacker)
        min_persistent_ttl = Uint32.unpack(unpacker)
        persistent_rent_rate_denominator = Int64.unpack(unpacker)
        temp_rent_rate_denominator = Int64.unpack(unpacker)
        max_entries_to_archive = Uint32.unpack(unpacker)
        bucket_list_size_window_sample_size = Uint32.unpack(unpacker)
        bucket_list_window_sample_period = Uint32.unpack(unpacker)
        eviction_scan_size = Uint32.unpack(unpacker)
        starting_eviction_scan_level = Uint32.unpack(unpacker)
        return cls(
            max_entry_ttl=max_entry_ttl,
            min_temporary_ttl=min_temporary_ttl,
            min_persistent_ttl=min_persistent_ttl,
            persistent_rent_rate_denominator=persistent_rent_rate_denominator,
            temp_rent_rate_denominator=temp_rent_rate_denominator,
            max_entries_to_archive=max_entries_to_archive,
            bucket_list_size_window_sample_size=bucket_list_size_window_sample_size,
            bucket_list_window_sample_period=bucket_list_window_sample_period,
            eviction_scan_size=eviction_scan_size,
            starting_eviction_scan_level=starting_eviction_scan_level,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> StateArchivalSettings:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> StateArchivalSettings:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.max_entry_ttl,
                self.min_temporary_ttl,
                self.min_persistent_ttl,
                self.persistent_rent_rate_denominator,
                self.temp_rent_rate_denominator,
                self.max_entries_to_archive,
                self.bucket_list_size_window_sample_size,
                self.bucket_list_window_sample_period,
                self.eviction_scan_size,
                self.starting_eviction_scan_level,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.max_entry_ttl == other.max_entry_ttl
            and self.min_temporary_ttl == other.min_temporary_ttl
            and self.min_persistent_ttl == other.min_persistent_ttl
            and self.persistent_rent_rate_denominator
            == other.persistent_rent_rate_denominator
            and self.temp_rent_rate_denominator == other.temp_rent_rate_denominator
            and self.max_entries_to_archive == other.max_entries_to_archive
            and self.bucket_list_size_window_sample_size
            == other.bucket_list_size_window_sample_size
            and self.bucket_list_window_sample_period
            == other.bucket_list_window_sample_period
            and self.eviction_scan_size == other.eviction_scan_size
            and self.starting_eviction_scan_level == other.starting_eviction_scan_level
        )

    def __repr__(self):
        out = [
            f"max_entry_ttl={self.max_entry_ttl}",
            f"min_temporary_ttl={self.min_temporary_ttl}",
            f"min_persistent_ttl={self.min_persistent_ttl}",
            f"persistent_rent_rate_denominator={self.persistent_rent_rate_denominator}",
            f"temp_rent_rate_denominator={self.temp_rent_rate_denominator}",
            f"max_entries_to_archive={self.max_entries_to_archive}",
            f"bucket_list_size_window_sample_size={self.bucket_list_size_window_sample_size}",
            f"bucket_list_window_sample_period={self.bucket_list_window_sample_period}",
            f"eviction_scan_size={self.eviction_scan_size}",
            f"starting_eviction_scan_level={self.starting_eviction_scan_level}",
        ]
        return f"<StateArchivalSettings [{', '.join(out)}]>"
